"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
from datetime import date, timedelta
from sqlalchemy.orm.session import Session
import pytest
from math import ceil
from decimal import Decimal
from copy import deepcopy

from biweeklybudget.interest import (
    InterestHelper,
    _InterestCalculation, AdbCompoundedDaily, SimpleInterest,
    _BillingPeriod,
    _MinPaymentFormula, MinPaymentAmEx, MinPaymentCiti, MinPaymentDiscover,
    _PayoffMethod, MinPaymentMethod, FixedPaymentMethod,
    LowestBalanceFirstMethod, HighestBalanceFirstMethod,
    LowestInterestRateFirstMethod, HighestInterestRateFirstMethod,
    calculate_payoffs, CCStatement,
    INTEREST_CALCULATION_NAMES, MIN_PAYMENT_FORMULA_NAMES,
    PAYOFF_METHOD_NAMES
)
from biweeklybudget.utils import dtnow
from biweeklybudget.models.account import Account, AcctType
from biweeklybudget.tests.unit_helpers import binexp_to_dict
from biweeklybudget.models.account_balance import AccountBalance


# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    PY34PLUS = False
    from mock import Mock, PropertyMock, call, patch
else:
    PY34PLUS = True
    from unittest.mock import Mock, PropertyMock, call, patch


pbm = 'biweeklybudget.interest'
pb = '%s.InterestHelper' % pbm


def pctdiff(a, b):
    """Return the percent difference of two Decimals, as a Decimal percent"""
    return abs(a - b) / abs(a)


class FixedInterest(_InterestCalculation):
    """
    Test class for fixed interest amount.
    """

    #: Human-readable string name of the interest calculation type.
    description = 'Test class for fixed interest amount.'

    def calculate(self, principal, first_d, last_d, transactions={}):
        """
        Calculate compound interest for the specified principal.

        :param principal: balance at beginning of statement period
        :type principal: decimal.Decimal
        :param first_d: date of beginning of statement period
        :type first_d: datetime.date
        :param last_d: last date of statement period
        :type last_d: datetime.date
        :param transactions: dict of datetime.date to float amount adjust
          the balance by on the specified dates.
        :type transactions: dict
        :return: dict describing the result: end_balance (float),
          interest_paid (float)
        :rtype: dict
        """
        num_days = 0

        d = first_d
        while d <= last_d:
            num_days += 1
            if d in transactions:
                principal += transactions[d]
            d += timedelta(days=1)
        return {
            'interest_paid': self.apr,
            'end_balance': principal + self.apr
        }


class TestInterestHelper(object):

    def setup_method(self):
        bal3 = Mock(
            spec_set=AccountBalance,
            ledger=Decimal('-952.06'),
            ledger_date=(dtnow() - timedelta(hours=13))
        )
        bal4 = Mock(
            spec_set=AccountBalance,
            ledger=Decimal('-5498.65'),
            ledger_date=(dtnow() - timedelta(hours=36))
        )
        self.accts = {
            3: Mock(
                spec_set=Account,
                id=3,
                description='First Credit Card, limit 2000',
                name='CreditOne',
                ofx_cat_memo_to_name=False,
                acct_type=AcctType.Credit,
                credit_limit=Decimal('2000.00'),
                is_active=True,
                apr=Decimal('0.0100'),
                negate_ofx_amounts=True,
                interest_class_name='AdbCompoundedDaily',
                min_payment_class_name='MinPaymentAmEx',
                balance=bal3,
                effective_apr=Decimal('0.0100'),
                last_interest_charge=Decimal('0.8089')
            ),
            4: Mock(
                spec_set=Account,
                id=4,
                description='Credit 2 limit 5500',
                name='CreditTwo',
                ofx_cat_memo_to_name=False,
                ofxgetter_config_json='',
                vault_creds_path='/foo/bar',
                acct_type=AcctType.Credit,
                credit_limit=Decimal('5500'),
                is_active=True,
                apr=Decimal('0.1000'),
                interest_class_name='AdbCompoundedDaily',
                min_payment_class_name='MinPaymentDiscover',
                balance=bal4,
                effective_apr=Decimal('0.1000'),
                last_interest_charge=Decimal('46.9061')
            )
        }
        self.mock_sess = Mock(spec_set=Session)
        self.mock_sess.query.return_value.filter.return_value.all.\
            return_value = self.accts.values()
        self.cls = InterestHelper(self.mock_sess)

    def test_init(self):
        assert self.cls._increases == {}
        assert self.cls._onetimes == {}
        assert self.mock_sess.mock_calls[0] == call.query(Account)
        kall = self.mock_sess.mock_calls[1]
        assert kall[0] == 'query().filter'
        assert len(kall[1]) == 2
        assert binexp_to_dict(kall[1][0]) == binexp_to_dict(
            Account.acct_type.__eq__(AcctType.Credit)
        )
        assert str(kall[1][1]) == 'accounts.is_active = true'
        assert self.mock_sess.mock_calls[2] == call.query().filter().all()
        assert self.cls.accounts == {
            3: self.accts[3],
            4: self.accts[4]
        }
        assert len(self.cls._statements) == 2
        s3 = self.cls._statements[3]
        assert isinstance(s3, CCStatement)
        assert isinstance(s3._billing_period, _BillingPeriod)
        assert s3._billing_period._period_for_date == date(2017, 7, 27)
        assert isinstance(s3._interest_cls, AdbCompoundedDaily)
        assert s3._interest_cls.apr == Decimal('0.0100')
        assert isinstance(s3._min_pay_cls, MinPaymentAmEx)
        assert s3._orig_principal == Decimal('952.06')
        assert s3._min_pay is None
        assert s3._transactions == {}
        assert s3._principal == Decimal('952.06')
        s4 = self.cls._statements[4]
        assert isinstance(s4, CCStatement)
        assert isinstance(s4._billing_period, _BillingPeriod)
        assert s4._billing_period._period_for_date == date(2017, 7, 26)
        assert isinstance(s4._interest_cls, AdbCompoundedDaily)
        assert s4._interest_cls.apr == Decimal('0.1000')
        assert isinstance(s4._min_pay_cls, MinPaymentDiscover)
        assert s4._orig_principal == Decimal('5498.65')
        assert s4._min_pay is None
        assert s4._transactions == {}
        assert s4._principal == Decimal('5498.65')

    def test_init_increases_onetimes(self):
        cls = InterestHelper(
            self.mock_sess,
            increases={'foo': 'bar'},
            onetimes={'baz': 'blam'}
        )
        assert cls._increases == {'foo': 'bar'}
        assert cls._onetimes == {'baz': 'blam'}

    def test_min_payments(self):
        res = self.cls.min_payments
        assert res == {
            3: Decimal('35'),
            4: Decimal('109.9730')
        }

    def test_calculate_payoffs(self):
        pm1 = Mock()
        pm2 = Mock()
        pm3 = Mock()
        type(pm3).show_in_ui = False
        meth_names = {
            'PM1': {
                'description': 'pm1desc',
                'doc': 'pm1doc',
                'cls': pm1
            },
            'PM2': {
                'description': 'pm2desc',
                'doc': 'pm2doc',
                'cls': pm2
            },
            'PM3': {
                'description': 'pm3desc',
                'doc': 'pm3doc',
                'cls': pm3
            }
        }
        with patch('%s.PAYOFF_METHOD_NAMES' % pbm, meth_names):
            with patch('%s._calc_payoff_method' % pb) as mock_cpm:
                mock_cpm.side_effect = ['res1', 'res2']
                res = self.cls.calculate_payoffs()
        assert res == {
            'PM1': {
                'description': 'pm1desc',
                'doc': 'pm1doc',
                'results': 'res1'
            },
            'PM2': {
                'description': 'pm2desc',
                'doc': 'pm2doc',
                'results': 'res2'
            }
        }
        assert mock_cpm.mock_calls == [
            call(pm1.return_value),
            call(pm2.return_value)
        ]

    def test_calculate_payoff_method(self):
        mock_m = Mock()
        with patch('%s.calculate_payoffs' % pbm) as mock_calc:
            mock_calc.return_value = [
                (28, Decimal('963.2130700030116938658705389'), Decimal('12.3')),
                (164, Decimal('8764.660910733671904414120065'), Decimal('4.56'))
            ]
            res = self.cls._calc_payoff_method(mock_m)
        assert res == {
            3: {
                'payoff_months': 28,
                'total_payments': Decimal('963.2130700030116938658705389'),
                'total_interest': Decimal('11.1530700030116938658705389'),
                'next_payment': Decimal('12.3')
            },
            4: {
                'payoff_months': 164,
                'total_payments': Decimal('8764.660910733671904414120065'),
                'total_interest': Decimal('3266.010910733671904414120065'),
                'next_payment': Decimal('4.56')
            }
        }


class TestInterestCalculation(object):

    def test_init(self):
        cls = _InterestCalculation(Decimal('1.23'))
        assert cls._apr == Decimal('1.23')

    def test_description(self):
        cls = _InterestCalculation(Decimal('0.1000'))
        assert hasattr(cls, 'description')

    def test_repr(self):
        cls = _InterestCalculation(Decimal('0.1000'))
        assert str(cls) == '<_InterestCalculation(decimal.Decimal(\'0.1000\'))>'


class TestAdbCompoundedDaily(object):

    def test_description(self):
        cls = AdbCompoundedDaily(Decimal('0.1000'))
        assert hasattr(cls, 'description')

    def test_repr(self):
        cls = AdbCompoundedDaily(Decimal('0.1000'))
        assert str(cls) == '<AdbCompoundedDaily(decimal.Decimal(\'0.1000\'))>'

    def test_calculate(self):
        cls = AdbCompoundedDaily(Decimal('0.1000'))
        res = cls.calculate(
            Decimal('100.00'),
            date(2017, 1, 1),
            (date(2017, 1, 1) + timedelta(days=365))
        )
        assert res == {
            'end_balance': Decimal('110.5487464695276899243379188'),
            'interest_paid': Decimal('10.54874567794529906536521621')
        }

    def test_calculate_transactions(self):
        cls = AdbCompoundedDaily(Decimal('0.1000'))
        end_d = date(2017, 1, 1) + timedelta(days=365)
        res = cls.calculate(
            Decimal('100.00'),
            date(2017, 1, 1),
            end_d,
            transactions={
                date(2017, 6, 1): Decimal('-50.00'),
                (end_d - timedelta(days=1)): Decimal('50.00')
            }
        )
        assert res == {
            'end_balance': Decimal('107.5420752170470026908058809'),
            'interest_paid': Decimal('7.542074651086492634526111808')
        }


class TestSimpleInterest(object):

    def test_description(self):
        cls = SimpleInterest(Decimal('0.1200'))
        assert hasattr(cls, 'description')

    def test_repr(self):
        cls = SimpleInterest(Decimal('0.1200'))
        assert str(cls) == '<SimpleInterest(decimal.Decimal(\'0.1200\'))>'

    def test_calculate(self):
        cls = SimpleInterest(Decimal('0.0100'))
        res = cls.calculate(
            Decimal('100.00'),
            date(2017, 1, 1),
            (date(2017, 1, 1) + timedelta(days=364))
        )
        assert res == {
            'end_balance': Decimal('101.0'),
            'interest_paid': Decimal('1.0')
        }

    def test_calculate_transactions(self):
        cls = SimpleInterest(Decimal('0.0100'))
        end_d = date(2017, 1, 1) + timedelta(days=364)
        res = cls.calculate(
            Decimal('100.00'),
            date(2017, 1, 1),
            end_d,
            transactions={
                date(2017, 6, 1): Decimal('-50.00')
            }
        )
        assert res == {
            'end_balance': Decimal('50.5'),
            'interest_paid': Decimal('0.5')
        }


class TestBillingPeriod(object):

    def test_init(self):
        cls = _BillingPeriod(date(2017, 1, 20))
        assert cls._end_date == date(2017, 1, 31)
        assert cls._start_date == date(2017, 1, 1)
        cls = _BillingPeriod(date(2017, 1, 5))
        assert cls._end_date == date(2016, 12, 31)
        assert cls._start_date == date(2016, 12, 1)

    def test_description(self):
        cls = _BillingPeriod(date(2017, 1, 20))
        assert hasattr(cls, 'description')

    def test_start_date(self):
        cls = _BillingPeriod(date(2017, 1, 20))
        cls._start_date = 'foo'
        assert cls.start_date == 'foo'

    def test_end_date(self):
        cls = _BillingPeriod(date(2017, 1, 20))
        cls._end_date = 'bar'
        assert cls.end_date == 'bar'

    def test_payment_date(self):
        cls = _BillingPeriod(date(2017, 1, 20))
        cls._start_date = date(2017, 1, 1)
        cls._end_date = date(2017, 1, 31)
        assert cls.payment_date == date(2017, 1, 16)

    def test_next(self):
        cls = _BillingPeriod(date(2017, 1, 1))
        cls._start_date = date(2017, 1, 1)
        cls._end_date = date(2017, 1, 31)
        n = cls.next_period
        assert n.start_date == date(2017, 2, 1)
        assert n.end_date == date(2017, 2, 28)

    def test_previous(self):
        cls = _BillingPeriod(date(2017, 1, 1))
        cls._start_date = date(2017, 1, 1)
        cls._end_date = date(2017, 1, 31)
        n = cls.prev_period
        assert n.start_date == date(2016, 12, 1)
        assert n.end_date == date(2016, 12, 31)


class TestMinPaymentFormula(object):

    def test_init(self):
        res = _MinPaymentFormula()
        assert isinstance(res, _MinPaymentFormula)

    def test_type(self):
        cls = _MinPaymentFormula()
        assert hasattr(cls, 'description')


class TestMinPaymentAmEx(object):

    def test_init(self):
        cls = MinPaymentAmEx()
        assert isinstance(cls, _MinPaymentFormula)

    def test_type(self):
        cls = MinPaymentAmEx()
        assert hasattr(cls, 'description')

    def test_calculate_over_35(self):
        cls = MinPaymentAmEx()
        res = cls.calculate(Decimal('10000.00'), Decimal('22.00'))
        assert res == Decimal('122.00')

    def test_calculate_under_35(self):
        cls = MinPaymentAmEx()
        res = cls.calculate(Decimal('100.00'), Decimal('22.00'))
        assert res == Decimal('35.00')


class TestMinPaymentDiscover(object):

    def test_init(self):
        cls = MinPaymentDiscover()
        assert isinstance(cls, _MinPaymentFormula)

    def test_type(self):
        cls = MinPaymentDiscover()
        assert hasattr(cls, 'description')

    def test_calculate_under_35(self):
        cls = MinPaymentDiscover()
        res = cls.calculate(Decimal('1.00'), Decimal('2.00'))
        assert res == Decimal('35.00')

    def test_calculate_2_percent(self):
        cls = MinPaymentDiscover()
        res = cls.calculate(Decimal('10000.00'), Decimal('2.00'))
        assert res == Decimal('200.00')

    def test_calculate_interest_plus_20(self):
        cls = MinPaymentDiscover()
        res = cls.calculate(Decimal('1.00'), Decimal('20.00'))
        assert res == Decimal('40.00')


class TestMinPaymentCiti(object):

    def test_init(self):
        cls = MinPaymentCiti()
        assert isinstance(cls, _MinPaymentFormula)

    def test_type(self):
        cls = MinPaymentCiti()
        assert hasattr(cls, 'description')

    def test_1point5_percent(self):
        cls = MinPaymentCiti()
        res = cls.calculate(Decimal('10000.00'), Decimal('2.00'))
        assert res == Decimal('150.00')

    def test_1percent_plus_interest(self):
        cls = MinPaymentCiti()
        res = cls.calculate(Decimal('10000.00'), Decimal('300.00'))
        assert res == Decimal('400.00')

    def test_25(self):
        cls = MinPaymentCiti()
        res = cls.calculate(Decimal('1.00'), Decimal('2.00'))
        assert res == Decimal('25.00')


class TestPayoffMethod(object):

    def test_init(self):
        cls = _PayoffMethod(Decimal('1.23'))
        assert isinstance(cls, _PayoffMethod)
        assert cls._max_total == Decimal('1.23')

    def test_description(self):
        cls = _PayoffMethod()
        assert hasattr(cls, 'description')

    def test_repr(self):
        cls = _PayoffMethod()
        assert 'PayoffMethod' in cls.__repr__()


class TestMinPaymentMethod(object):

    def test_init(self):
        cls = MinPaymentMethod()
        assert isinstance(cls, _PayoffMethod)

    def test_description(self):
        cls = MinPaymentMethod()
        assert hasattr(cls, 'description')

    def test_find_payments(self):
        cls = MinPaymentMethod()
        s1 = Mock(spec_set=CCStatement)
        type(s1).minimum_payment = PropertyMock(return_value=Decimal('1.11'))
        s2 = Mock(spec_set=CCStatement)
        type(s2).minimum_payment = PropertyMock(return_value=Decimal('2.22'))
        s3 = Mock(spec_set=CCStatement)
        type(s3).minimum_payment = PropertyMock(return_value=Decimal('3.33'))
        assert cls.find_payments([s1, s2, s3]) == [
            Decimal('1.11'), Decimal('2.22'), Decimal('3.33')
        ]


class TestFixedPaymentMethod(object):

    def test_init(self):
        cls = FixedPaymentMethod(Decimal('12.34'))
        assert isinstance(cls, _PayoffMethod)
        assert cls._max_total == Decimal('12.34')

    def test_description(self):
        cls = FixedPaymentMethod(Decimal('12.34'))
        assert hasattr(cls, 'description')

    def test_find_payments(self):
        cls = FixedPaymentMethod(Decimal('12.34'))
        s1 = Mock(spec_set=CCStatement)
        type(s1).minimum_payment = PropertyMock(return_value=Decimal('1.11'))
        s2 = Mock(spec_set=CCStatement)
        type(s2).minimum_payment = PropertyMock(return_value=Decimal('2.22'))
        s3 = Mock(spec_set=CCStatement)
        type(s3).minimum_payment = PropertyMock(return_value=Decimal('3.33'))
        assert cls.find_payments([s1, s2, s3]) == [
            Decimal('12.34'), Decimal('12.34'), Decimal('12.34')
        ]


class TestLowestBalanceFirstMethod(object):

    def test_init(self):
        cls = LowestBalanceFirstMethod(Decimal('100.00'))
        assert isinstance(cls, _PayoffMethod)
        assert cls._max_total == Decimal('100.00')

    def test_description(self):
        cls = LowestBalanceFirstMethod(Decimal('100.00'))
        assert hasattr(cls, 'description')

    def test_find_payments(self):
        cls = LowestBalanceFirstMethod(Decimal('100.00'))
        s1 = Mock(spec_set=CCStatement)
        type(s1).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s1).principal = PropertyMock(return_value=Decimal('10.00'))
        s2 = Mock(spec_set=CCStatement)
        type(s2).minimum_payment = PropertyMock(return_value=Decimal('5.00'))
        type(s2).principal = PropertyMock(return_value=Decimal('25.00'))
        s3 = Mock(spec_set=CCStatement)
        type(s3).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s3).principal = PropertyMock(return_value=Decimal('1234.56'))
        s4 = Mock(spec_set=CCStatement)
        type(s4).minimum_payment = PropertyMock(return_value=Decimal('7.00'))
        type(s4).principal = PropertyMock(return_value=Decimal('3.00'))
        assert cls.find_payments([s1, s2, s3, s4]) == [
            Decimal('2.00'),
            Decimal('5.00'),
            Decimal('2.00'),
            Decimal('91.00')
        ]

    def test_find_payments_total_too_low(self):
        cls = LowestBalanceFirstMethod(Decimal('3.00'))
        s1 = Mock(spec_set=CCStatement)
        type(s1).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s1).principal = PropertyMock(return_value=Decimal('10.00'))
        s2 = Mock(spec_set=CCStatement)
        type(s2).minimum_payment = PropertyMock(return_value=Decimal('5.00'))
        type(s2).principal = PropertyMock(return_value=Decimal('25.00'))
        s3 = Mock(spec_set=CCStatement)
        type(s3).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s3).principal = PropertyMock(return_value=Decimal('1234.56'))
        s4 = Mock(spec_set=CCStatement)
        type(s4).minimum_payment = PropertyMock(return_value=Decimal('7.00'))
        type(s4).principal = PropertyMock(return_value=Decimal('3.00'))
        with pytest.raises(TypeError):
            cls.find_payments([s1, s2, s3, s4])


class TestHighestBalanceFirstMethod(object):

    def test_init(self):
        cls = HighestBalanceFirstMethod(Decimal('100.00'))
        assert isinstance(cls, _PayoffMethod)
        assert cls._max_total == Decimal('100.00')

    def test_description(self):
        cls = HighestBalanceFirstMethod(Decimal('100.00'))
        assert hasattr(cls, 'description')

    def test_find_payments(self):
        cls = HighestBalanceFirstMethod(Decimal('100.00'))
        s1 = Mock(spec_set=CCStatement)
        type(s1).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s1).principal = PropertyMock(return_value=Decimal('10.00'))
        s2 = Mock(spec_set=CCStatement)
        type(s2).minimum_payment = PropertyMock(return_value=Decimal('5.00'))
        type(s2).principal = PropertyMock(return_value=Decimal('25.00'))
        s3 = Mock(spec_set=CCStatement)
        type(s3).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s3).principal = PropertyMock(return_value=Decimal('1234.56'))
        s4 = Mock(spec_set=CCStatement)
        type(s4).minimum_payment = PropertyMock(return_value=Decimal('7.00'))
        type(s4).principal = PropertyMock(return_value=Decimal('3.00'))
        assert cls.find_payments([s1, s2, s3, s4]) == [
            Decimal('2.00'),
            Decimal('5.00'),
            Decimal('86.00'),
            Decimal('7.00')
        ]

    def test_find_payments_total_too_low(self):
        cls = HighestBalanceFirstMethod(Decimal('3.00'))
        s1 = Mock(spec_set=CCStatement)
        type(s1).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s1).principal = PropertyMock(return_value=Decimal('10.00'))
        s2 = Mock(spec_set=CCStatement)
        type(s2).minimum_payment = PropertyMock(return_value=Decimal('5.00'))
        type(s2).principal = PropertyMock(return_value=Decimal('25.00'))
        s3 = Mock(spec_set=CCStatement)
        type(s3).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s3).principal = PropertyMock(return_value=Decimal('1234.56'))
        s4 = Mock(spec_set=CCStatement)
        type(s4).minimum_payment = PropertyMock(return_value=Decimal('7.00'))
        type(s4).principal = PropertyMock(return_value=Decimal('3.00'))
        with pytest.raises(TypeError):
            cls.find_payments([s1, s2, s3, s4])


class TestHighestInterestRateFirstMethod(object):

    def test_init(self):
        cls = HighestInterestRateFirstMethod(Decimal('100.00'))
        assert isinstance(cls, _PayoffMethod)
        assert cls._max_total == Decimal('100.00')

    def test_description(self):
        cls = HighestInterestRateFirstMethod(Decimal('100.00'))
        assert hasattr(cls, 'description')

    def test_find_payments(self):
        cls = HighestInterestRateFirstMethod(Decimal('100.00'))
        s1 = Mock(spec_set=CCStatement)
        type(s1).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s1).apr = PropertyMock(return_value=Decimal('0.0100'))
        type(s1).principal = PropertyMock(return_value=Decimal('10.00'))
        s2 = Mock(spec_set=CCStatement)
        type(s2).minimum_payment = PropertyMock(return_value=Decimal('5.00'))
        type(s2).apr = PropertyMock(return_value=Decimal('0.0200'))
        type(s2).principal = PropertyMock(return_value=Decimal('25.00'))
        s3 = Mock(spec_set=CCStatement)
        type(s3).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s3).apr = PropertyMock(return_value=Decimal('0.0800'))
        type(s3).principal = PropertyMock(return_value=Decimal('1234.56'))
        s4 = Mock(spec_set=CCStatement)
        type(s4).minimum_payment = PropertyMock(return_value=Decimal('7.00'))
        type(s4).apr = PropertyMock(return_value=Decimal('0.0300'))
        type(s4).principal = PropertyMock(return_value=Decimal('3.00'))
        assert cls.find_payments([s1, s2, s3, s4]) == [
            Decimal('2.00'),
            Decimal('5.00'),
            Decimal('86.00'),
            Decimal('7.00')
        ]

    def test_find_payments_total_too_low(self):
        cls = HighestInterestRateFirstMethod(Decimal('3.00'))
        s1 = Mock(spec_set=CCStatement)
        type(s1).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s1).principal = PropertyMock(return_value=Decimal('10.00'))
        s2 = Mock(spec_set=CCStatement)
        type(s2).minimum_payment = PropertyMock(return_value=Decimal('5.00'))
        type(s2).principal = PropertyMock(return_value=Decimal('25.00'))
        s3 = Mock(spec_set=CCStatement)
        type(s3).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s3).principal = PropertyMock(return_value=Decimal('1234.56'))
        s4 = Mock(spec_set=CCStatement)
        type(s4).minimum_payment = PropertyMock(return_value=Decimal('7.00'))
        type(s4).principal = PropertyMock(return_value=Decimal('3.00'))
        with pytest.raises(TypeError):
            cls.find_payments([s1, s2, s3, s4])


class TestLowestInterestRateFirstMethod(object):

    def test_init(self):
        cls = LowestInterestRateFirstMethod(Decimal('100.00'))
        assert isinstance(cls, _PayoffMethod)
        assert cls._max_total == Decimal('100.00')

    def test_description(self):
        cls = LowestInterestRateFirstMethod(Decimal('100.00'))
        assert hasattr(cls, 'description')

    def test_find_payments(self):
        cls = LowestInterestRateFirstMethod(Decimal('100.00'))
        s1 = Mock(spec_set=CCStatement)
        type(s1).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s1).apr = PropertyMock(return_value=Decimal('0.0100'))
        type(s1).principal = PropertyMock(return_value=Decimal('10.00'))
        s2 = Mock(spec_set=CCStatement)
        type(s2).minimum_payment = PropertyMock(return_value=Decimal('5.00'))
        type(s2).apr = PropertyMock(return_value=Decimal('0.0200'))
        type(s2).principal = PropertyMock(return_value=Decimal('25.00'))
        s3 = Mock(spec_set=CCStatement)
        type(s3).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s3).apr = PropertyMock(return_value=Decimal('0.0800'))
        type(s3).principal = PropertyMock(return_value=Decimal('1234.56'))
        s4 = Mock(spec_set=CCStatement)
        type(s4).minimum_payment = PropertyMock(return_value=Decimal('7.00'))
        type(s4).apr = PropertyMock(return_value=Decimal('0.0300'))
        type(s4).principal = PropertyMock(return_value=Decimal('3.00'))
        assert cls.find_payments([s1, s2, s3, s4]) == [
            Decimal('86.00'),
            Decimal('5.00'),
            Decimal('2.00'),
            Decimal('7.00')
        ]

    def test_find_payments_total_too_low(self):
        cls = LowestInterestRateFirstMethod(Decimal('3.00'))
        s1 = Mock(spec_set=CCStatement)
        type(s1).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s1).principal = PropertyMock(return_value=Decimal('10.00'))
        s2 = Mock(spec_set=CCStatement)
        type(s2).minimum_payment = PropertyMock(return_value=Decimal('5.00'))
        type(s2).principal = PropertyMock(return_value=Decimal('25.00'))
        s3 = Mock(spec_set=CCStatement)
        type(s3).minimum_payment = PropertyMock(return_value=Decimal('2.00'))
        type(s3).principal = PropertyMock(return_value=Decimal('1234.56'))
        s4 = Mock(spec_set=CCStatement)
        type(s4).minimum_payment = PropertyMock(return_value=Decimal('7.00'))
        type(s4).principal = PropertyMock(return_value=Decimal('3.00'))
        with pytest.raises(TypeError):
            cls.find_payments([s1, s2, s3, s4])


class TestCCStatement(object):

    def test_init(self):
        b = Mock(spec_set=_BillingPeriod)
        i = Mock(spec_set=_InterestCalculation)
        p = Mock(spec_set=_MinPaymentFormula)
        cls = CCStatement(
            i, Decimal('1.23'), p, b,
            end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
        )
        assert cls._billing_period == b
        assert cls._interest_cls == i
        assert cls._min_pay_cls == p
        assert cls._orig_principal == Decimal('1.23')
        assert cls._min_pay is None
        assert cls._transactions == {}
        assert cls._principal == Decimal('1.50')
        assert cls._interest_amt == Decimal('0.27')
        assert i.mock_calls == []

    def test_init_end_none(self):
        b = Mock(spec_set=_BillingPeriod)
        type(b).start_date = PropertyMock(return_value=date(2017, 1, 1))
        type(b).end_date = PropertyMock(return_value=date(2017, 1, 29))
        i = Mock(spec_set=_InterestCalculation)
        i.calculate.return_value = {
            'end_balance': Decimal('100.11'),
            'interest_paid': Decimal('23.45')
        }
        p = Mock(spec_set=_MinPaymentFormula)
        cls = CCStatement(
            i, Decimal('1.23'), p, b,
            interest_amt=Decimal('0.27')
        )
        assert cls._billing_period == b
        assert cls._interest_cls == i
        assert cls._min_pay_cls == p
        assert cls._orig_principal == Decimal('1.23')
        assert cls._min_pay is None
        assert cls._transactions == {}
        assert cls._principal == Decimal('100.11')
        assert cls._interest_amt == Decimal('0.27')
        assert i.mock_calls == [
            call.calculate(
                Decimal('1.23'),
                date(2017, 1, 1),
                date(2017, 1, 29),
                {}
            )
        ]
        assert 'CCStatement' in str(cls)

    def test_init_int_none(self):
        b = Mock(spec_set=_BillingPeriod)
        type(b).start_date = PropertyMock(return_value=date(2017, 1, 1))
        type(b).end_date = PropertyMock(return_value=date(2017, 1, 29))
        i = Mock(spec_set=_InterestCalculation)
        i.calculate.return_value = {
            'end_balance': Decimal('100.11'),
            'interest_paid': Decimal('23.45')
        }
        p = Mock(spec_set=_MinPaymentFormula)
        cls = CCStatement(
            i, Decimal('1.23'), p, b,
            end_balance=Decimal('1.50')
        )
        assert cls._billing_period == b
        assert cls._interest_cls == i
        assert cls._min_pay_cls == p
        assert cls._orig_principal == Decimal('1.23')
        assert cls._min_pay is None
        assert cls._transactions == {}
        assert cls._principal == Decimal('1.50')
        assert cls._interest_amt == Decimal('23.45')
        assert i.mock_calls == [
            call.calculate(
                Decimal('1.23'),
                date(2017, 1, 1),
                date(2017, 1, 29),
                {}
            )
        ]

    def test_init_interest_wrong_class(self):
        b = Mock(spec_set=_BillingPeriod)
        i = None
        p = Mock(spec_set=_MinPaymentFormula)
        with pytest.raises(TypeError):
            CCStatement(
                i, Decimal('1.23'), p, b,
                end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
            )

    def test_init_payment_wrong_class(self):
        b = Mock(spec_set=_BillingPeriod)
        i = Mock(spec_set=_InterestCalculation)
        p = None
        with pytest.raises(TypeError):
            CCStatement(
                i, Decimal('1.23'), p, b,
                end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
            )

    def test_init_billing_wrong_class(self):
        b = None
        i = Mock(spec_set=_InterestCalculation)
        p = Mock(spec_set=_MinPaymentFormula)
        with pytest.raises(TypeError):
            CCStatement(
                i, Decimal('1.23'), p, b,
                end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
            )

    def test_principal(self):
        b = Mock(spec_set=_BillingPeriod)
        i = Mock(spec_set=_InterestCalculation)
        p = Mock(spec_set=_MinPaymentFormula)
        cls = CCStatement(
            i, Decimal('1.23'), p, b,
            end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
        )
        assert cls._principal == Decimal('1.50')
        assert cls.principal == Decimal('1.50')

    def test_interest(self):
        b = Mock(spec_set=_BillingPeriod)
        i = Mock(spec_set=_InterestCalculation)
        p = Mock(spec_set=_MinPaymentFormula)
        cls = CCStatement(
            i, Decimal('1.23'), p, b,
            end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
        )
        assert cls._interest_amt == Decimal('0.27')
        assert cls.interest == Decimal('0.27')

    def test_start_date(self):
        b = Mock(spec_set=_BillingPeriod)
        type(b).start_date = PropertyMock(return_value=date(2017, 1, 1))
        type(b).end_date = PropertyMock(return_value=date(2017, 1, 29))
        i = Mock(spec_set=_InterestCalculation)
        p = Mock(spec_set=_MinPaymentFormula)
        cls = CCStatement(
            i, Decimal('1.23'), p, b,
            end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
        )
        assert cls.start_date == date(2017, 1, 1)

    def test_end_date(self):
        b = Mock(spec_set=_BillingPeriod)
        type(b).start_date = PropertyMock(return_value=date(2017, 1, 1))
        type(b).end_date = PropertyMock(return_value=date(2017, 1, 29))
        i = Mock(spec_set=_InterestCalculation)
        p = Mock(spec_set=_MinPaymentFormula)
        cls = CCStatement(
            i, Decimal('1.23'), p, b,
            end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
        )
        assert cls.end_date == date(2017, 1, 29)

    def test_apr(self):
        b = Mock(spec_set=_BillingPeriod)
        i = Mock(spec_set=_InterestCalculation)
        type(i).apr = PropertyMock(return_value=Decimal('0.1234'))
        p = Mock(spec_set=_MinPaymentFormula)
        cls = CCStatement(
            i, Decimal('1.23'), p, b,
            end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
        )
        assert cls.apr == Decimal('0.1234')

    def test_minimum_payment(self):
        b = Mock(spec_set=_BillingPeriod)
        i = Mock(spec_set=_InterestCalculation)
        p = Mock(spec_set=_MinPaymentFormula)
        p.calculate.return_value = Decimal('123.45')
        cls = CCStatement(
            i, Decimal('1.23'), p, b,
            end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
        )
        assert cls._min_pay_cls == p
        assert cls._principal == Decimal('1.50')
        assert cls._interest_amt == Decimal('0.27')
        assert cls.minimum_payment == Decimal('123.45')
        assert p.mock_calls == [
            call.calculate(Decimal('1.50'), Decimal('0.27'))
        ]

    def test_next_with_transactions(self):
        b = Mock(spec_set=_BillingPeriod)
        b_next = Mock(spec_set=_BillingPeriod)
        type(b).next_period = PropertyMock(return_value=b_next)
        i = Mock(spec_set=_InterestCalculation)
        p = Mock(spec_set=_MinPaymentFormula)
        cls = CCStatement(
            i, Decimal('1.23'), p, b,
            end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
        )
        mock_stmt = Mock(spec_set=CCStatement)
        with patch('biweeklybudget.interest.CCStatement', autospec=True) as m:
            m.return_value = mock_stmt
            res = cls.next_with_transactions({'foo': 'bar'})
        assert res == mock_stmt
        assert m.mock_calls == [
            call(
                i, Decimal('1.50'), p, b_next, transactions={'foo': 'bar'}
            )
        ]

    def test_pay(self):
        b = Mock(spec_set=_BillingPeriod)
        b_next = Mock(spec_set=_BillingPeriod)
        type(b_next).payment_date = PropertyMock(return_value=date(2017, 1, 15))
        type(b).next_period = PropertyMock(return_value=b_next)
        i = Mock(spec_set=_InterestCalculation)
        p = Mock(spec_set=_MinPaymentFormula)
        cls = CCStatement(
            i, Decimal('1.23'), p, b,
            end_balance=Decimal('1.50'), interest_amt=Decimal('0.27')
        )
        mock_stmt = Mock(spec_set=CCStatement)
        with patch(
            'biweeklybudget.interest.CCStatement.next_with_transactions',
            autospec=True
        ) as m:
            m.return_value = mock_stmt
            res = cls.pay(Decimal('98.76'))
        assert res == mock_stmt
        assert m.mock_calls == [
            call(cls, {date(2017, 1, 15): Decimal('98.76')})
        ]


class TestCalculatePayoffs(object):

    def test_simple(self):
        def se_interest(bal, *args, **kwargs):
            return {
                'end_balance': bal - Decimal('1.00'),
                'interest_paid': Decimal('1.00')
            }

        def se_minpay_A(*args):
            return Decimal('3.00')

        def se_minpay_B(*args):
            return Decimal('60.00')

        b1 = _BillingPeriod(date(2017, 1, 2))
        i1 = Mock(spec_set=_InterestCalculation)
        i1.calculate.side_effect = se_interest
        p1 = Mock(spec_set=_MinPaymentFormula)
        p1.calculate.side_effect = se_minpay_A
        s1 = CCStatement(
            i1, Decimal('10.00'), p1, b1,
            end_balance=Decimal('10.00'), interest_amt=Decimal('1.00')
        )
        p2 = Mock(spec_set=_MinPaymentFormula)
        p2.calculate.side_effect = se_minpay_B
        s2 = CCStatement(
            i1, Decimal('20.00'), p2, b1,
            end_balance=Decimal('20.00'), interest_amt=Decimal('2.00')
        )
        s3 = CCStatement(
            i1, Decimal('-1.00'), p2, b1,
            end_balance=Decimal('-1.00'), interest_amt=Decimal('2.00')
        )
        meth = MinPaymentMethod()
        res = calculate_payoffs(meth, [s1, s2, s3])
        assert res == [
            (8, Decimal('24.00'), Decimal('3.00')),
            (1, Decimal('20.00'), Decimal('20.00')),
            (0, Decimal('0.0'), Decimal('0.0'))
        ]


class TestModuleConstants(object):

    def test_interest(self):
        assert INTEREST_CALCULATION_NAMES == {
            'AdbCompoundedDaily': {
                'description': AdbCompoundedDaily.description,
                'doc': AdbCompoundedDaily.__doc__.strip(),
                'cls': AdbCompoundedDaily
            },
            'SimpleInterest': {
                'description': SimpleInterest.description,
                'doc': SimpleInterest.__doc__.strip(),
                'cls': SimpleInterest
            }
        }

    def test_min_payment(self):
        assert MIN_PAYMENT_FORMULA_NAMES == {
            'MinPaymentAmEx': {
                'description': MinPaymentAmEx.description,
                'doc': MinPaymentAmEx.__doc__.strip(),
                'cls': MinPaymentAmEx
            },
            'MinPaymentDiscover': {
                'description': MinPaymentDiscover.description,
                'doc': MinPaymentDiscover.__doc__.strip(),
                'cls': MinPaymentDiscover
            },
            'MinPaymentCiti': {
                'description': MinPaymentCiti.description,
                'doc': MinPaymentCiti.__doc__.strip(),
                'cls': MinPaymentCiti
            }
        }

    def test_payoff(self):
        assert PAYOFF_METHOD_NAMES == {
            'MinPaymentMethod': {
                'description': MinPaymentMethod.description,
                'doc': MinPaymentMethod.__doc__.strip(),
                'cls': MinPaymentMethod
            },
            'FixedPaymentMethod': {
                'description': FixedPaymentMethod.description,
                'doc': FixedPaymentMethod.__doc__.strip(),
                'cls': FixedPaymentMethod
            },
            'HighestBalanceFirstMethod': {
                'description': HighestBalanceFirstMethod.description,
                'doc': HighestBalanceFirstMethod.__doc__.strip(),
                'cls': HighestBalanceFirstMethod
            },
            'LowestBalanceFirstMethod': {
                'description': LowestBalanceFirstMethod.description,
                'doc': LowestBalanceFirstMethod.__doc__.strip(),
                'cls': LowestBalanceFirstMethod
            },
            'LowestInterestRateFirstMethod': {
                'description': LowestInterestRateFirstMethod.description,
                'doc': LowestInterestRateFirstMethod.__doc__.strip(),
                'cls': LowestInterestRateFirstMethod
            },
            'HighestInterestRateFirstMethod': {
                'description': HighestInterestRateFirstMethod.description,
                'doc': HighestInterestRateFirstMethod.__doc__.strip(),
                'cls': HighestInterestRateFirstMethod
            }
        }


class TestAcceptanceData(object):

    def setup_method(self):
        self.stmt_cc_one = CCStatement(
            AdbCompoundedDaily(Decimal('0.0100')),
            Decimal('952.06'),
            MinPaymentAmEx(),
            _BillingPeriod(date(2017, 7, 31)),
            transactions={},
            end_balance=Decimal('952.06'),
            interest_amt=Decimal('16.25')
        )
        self.stmt_cc_two = CCStatement(
            AdbCompoundedDaily(Decimal('0.1000')),
            Decimal('5498.65'),
            MinPaymentDiscover(),
            _BillingPeriod(date(2017, 7, 31)),
            transactions={},
            end_balance=Decimal('5498.65'),
            interest_amt=Decimal('28.53')
        )

    def test_cc_one_minimum(self):
        assert self.stmt_cc_one.minimum_payment == Decimal('35')

    def test_cc_two_minimum(self):
        assert self.stmt_cc_two.minimum_payment == Decimal('109.9730')

    def test_cc_one_pay_min(self):
        res = calculate_payoffs(
            MinPaymentMethod(),
            [self.stmt_cc_one]
        )
        assert res == [
            (28, Decimal('962.9988625702411101133192793'), Decimal('35'))
        ]

    def test_cc_two_pay_min(self):
        res = calculate_payoffs(
            MinPaymentMethod(),
            [self.stmt_cc_two]
        )
        assert res == [
            (
                162,
                Decimal('8664.861877369277471400473622'),
                Decimal('109.9730')
            )
        ]

    def test_cc_combined_minimum(self):
        assert (
            self.stmt_cc_one.minimum_payment + self.stmt_cc_two.minimum_payment
        ) == Decimal('144.9730')

    def test_combined_pay_min(self):
        res = calculate_payoffs(
            MinPaymentMethod(),
            [self.stmt_cc_one, self.stmt_cc_two]
        )
        assert res == [
            (28, Decimal('962.9988625702411101133192793'), Decimal('35')),
            (162, Decimal('8664.861877369277471400473622'), Decimal('109.9730'))
        ]

    def test_combined_pay_lowest_ir(self):
        res = calculate_payoffs(
            LowestInterestRateFirstMethod(Decimal('144.9730')),
            [self.stmt_cc_one, self.stmt_cc_two]
        )
        assert res == [
            (21, Decimal('960.9178327498502165965138131'), Decimal('35')),
            (56, Decimal('6988.237124948955044765363412'), Decimal('109.9730'))
        ]

    def test_combined_pay_lowest_bal(self):
        res = calculate_payoffs(
            LowestBalanceFirstMethod(Decimal('144.9730')),
            [self.stmt_cc_one, self.stmt_cc_two]
        )
        assert res == [
            (21, Decimal('960.9178327498502165965138131'), Decimal('35')),
            (56, Decimal('6988.237124948955044765363412'), Decimal('109.9730'))
        ]

    def test_combined_pay_highest_ir(self):
        res = calculate_payoffs(
            HighestInterestRateFirstMethod(Decimal('144.9730')),
            [self.stmt_cc_one, self.stmt_cc_two]
        )
        assert res == [
            (28, Decimal('962.9988625702411101133192793'), Decimal('35')),
            (55, Decimal('6956.345228060182432444990377'), Decimal('109.9730'))
        ]

    def test_combined_pay_highest_bal(self):
        res = calculate_payoffs(
            HighestBalanceFirstMethod(Decimal('144.9730')),
            [self.stmt_cc_one, self.stmt_cc_two]
        )
        assert res == [
            (28, Decimal('962.9988625702411101133192793'), Decimal('35')),
            (55, Decimal('6956.345228060182432444990377'), Decimal('109.9730'))
        ]


class TestSimpleData(object):

    def setup_method(self):
        self.mpm = Mock(spec_set=MinPaymentAmEx)
        self.mpm.calculate.return_value = Decimal('200.00')
        self.stmt_cc_one = CCStatement(
            FixedInterest(Decimal('10.00')),
            Decimal('1000.00'),
            self.mpm,
            _BillingPeriod(date(2017, 7, 3)),
            transactions={},
            end_balance=Decimal('1010.00'),
            interest_amt=Decimal('10.00')
        )
        self.stmt_cc_two = CCStatement(
            FixedInterest(Decimal('40.00')),
            Decimal('2000.00'),
            self.mpm,
            _BillingPeriod(date(2017, 7, 14)),
            transactions={},
            end_balance=Decimal('2040.00'),
            interest_amt=Decimal('40.00')
        )
        self.mpm2 = Mock(spec_set=MinPaymentAmEx)
        self.mpm2.calculate.return_value = Decimal('500.00')
        self.stmt_cc_three = CCStatement(
            FixedInterest(Decimal('100.00')),
            Decimal('10000.00'),
            self.mpm2,
            _BillingPeriod(date(2017, 7, 7)),
            transactions={},
            end_balance=Decimal('10100.00'),
            interest_amt=Decimal('100.00')
        )

    def test_cc_one_minimum(self):
        assert self.stmt_cc_one.minimum_payment == Decimal('200.0')

    def test_cc_two_minimum(self):
        assert self.stmt_cc_two.minimum_payment == Decimal('200.0')

    def test_cc_three_minimum(self):
        assert self.stmt_cc_three.minimum_payment == Decimal('500.0')

    def test_cc_one_pay_min(self):
        res = calculate_payoffs(
            MinPaymentMethod(),
            [self.stmt_cc_one]
        )
        assert res == [(6, Decimal('1060.0'), Decimal('200.00'))]

    def test_cc_two_pay_min(self):
        res = calculate_payoffs(
            MinPaymentMethod(),
            [self.stmt_cc_two]
        )
        assert res == [(13, Decimal('2520.0'), Decimal('200.00'))]

    def test_cc_three_pay_min(self):
        res = calculate_payoffs(
            MinPaymentMethod(),
            [self.stmt_cc_three]
        )
        assert res == [(25, Decimal('12500.0'), Decimal('500.00'))]

    def test_combined_pay_min(self):
        res = calculate_payoffs(
            MinPaymentMethod(),
            [self.stmt_cc_one, self.stmt_cc_two, self.stmt_cc_three]
        )
        assert res == [
            (6, Decimal('1060.0'), Decimal('200.00')),
            (13, Decimal('2520.0'), Decimal('200.00')),
            (25, Decimal('12500.0'), Decimal('500.00'))
        ]

    def test_lowest_balance_first(self):
        res = calculate_payoffs(
            LowestBalanceFirstMethod(Decimal('1000.0')),
            [self.stmt_cc_one, self.stmt_cc_two, self.stmt_cc_three]
        )
        assert res == [
            (4, Decimal('1040.0'), Decimal('300.00')),
            (7, Decimal('2280.0'), Decimal('200.00')),
            (15, Decimal('11500.0'), Decimal('500.00'))
        ]

    def test_highest_balance_first(self):
        res = calculate_payoffs(
            HighestBalanceFirstMethod(Decimal('1000.0')),
            [self.stmt_cc_one, self.stmt_cc_two, self.stmt_cc_three]
        )
        assert res == [
            (6, Decimal('1060.0'), Decimal('200.00')),
            (13, Decimal('2520.0'), Decimal('200.00')),
            (16, Decimal('11600.0'), Decimal('600.00'))
        ]

    def test_lowest_interest_rate_first(self):
        res = calculate_payoffs(
            LowestInterestRateFirstMethod(Decimal('1000.0')),
            [self.stmt_cc_one, self.stmt_cc_two, self.stmt_cc_three]
        )
        assert res == [
            (4, Decimal('1040.0'), Decimal('300.00')),
            (7, Decimal('2280.0'), Decimal('200.00')),
            (15, Decimal('11500.0'), Decimal('500.00'))
        ]

    def test_highest_interest_rate_first(self):
        res = calculate_payoffs(
            HighestInterestRateFirstMethod(Decimal('1000.0')),
            [self.stmt_cc_one, self.stmt_cc_two, self.stmt_cc_three]
        )
        assert res == [
            (6, Decimal('1060.0'), Decimal('200.00')),
            (13, Decimal('2520.0'), Decimal('200.00')),
            (16, Decimal('11600.0'), Decimal('600.00'))
        ]

    def test_lowest_balance_first_with_increases(self):
        res = calculate_payoffs(
            LowestBalanceFirstMethod(
                Decimal('1000.0'),
                increases={
                    date(2017, 8, 1): Decimal('1200.0'),  # bill 3
                    date(2018, 2, 1): Decimal('1500.0'),  # bill 9
                }
            ),
            [self.stmt_cc_one, self.stmt_cc_two, self.stmt_cc_three]
        )
        assert res == [
            (3, Decimal('1030.0'), Decimal('300.00')),
            (6, Decimal('2240.0'), Decimal('200.00')),
            (12, Decimal('11200.0'), Decimal('500.00'))
        ]

    def test_highest_balance_first_with_increases(self):
        res = calculate_payoffs(
            HighestBalanceFirstMethod(
                Decimal('1000.0'),
                increases={
                    date(2017, 8, 1): Decimal('1200.0'),  # bill 3
                    date(2018, 2, 1): Decimal('1500.0'),  # bill 9
                }
            ),
            [self.stmt_cc_one, self.stmt_cc_two, self.stmt_cc_three]
        )
        assert res == [
            (6, Decimal('1060.0'), Decimal('200.00')),
            (13, Decimal('2520.0'), Decimal('200.00')),
            (12, Decimal('11200.0'), Decimal('600.00'))
        ]

    def test_lowest_balance_first_with_onetimes(self):
        res = calculate_payoffs(
            LowestBalanceFirstMethod(
                Decimal('1000.0'),
                onetimes={
                    date(2017, 7, 1): Decimal('500.0'),  # bill 2
                    date(2017, 10, 1): Decimal('1000.0'),  # bill 5
                    date(2018, 4, 1): Decimal('3000.0')  # bill 11
                }
            ),
            [self.stmt_cc_one, self.stmt_cc_two, self.stmt_cc_three]
        )
        assert res == [
            (2, Decimal('1020.0'), Decimal('300.00')),
            (5, Decimal('2200.0'), Decimal('200.00')),
            (11, Decimal('11100.0'), Decimal('500.00'))
        ]

    def test_highest_balance_first_with_onetimes(self):
        res = calculate_payoffs(
            HighestBalanceFirstMethod(
                Decimal('1000.0'),
                onetimes={
                    date(2017, 7, 1): Decimal('500.0'),  # bill 2
                    date(2017, 10, 1): Decimal('1000.0'),  # bill 5
                    date(2018, 4, 1): Decimal('3000.0')  # bill 11
                }
            ),
            [self.stmt_cc_one, self.stmt_cc_two, self.stmt_cc_three]
        )
        assert res == [
            (6, Decimal('1060.0'), Decimal('200.00')),
            (12, Decimal('2480.0'), Decimal('200.00')),
            (11, Decimal('11100.0'), Decimal('600.00'))
        ]

    def test_lowest_balance_first_with_increases_and_onetimes(self):
        res = calculate_payoffs(
            LowestBalanceFirstMethod(
                Decimal('1000.0'),
                onetimes={
                    date(2017, 6, 1): Decimal('200.0'),  # bill 1
                    date(2017, 9, 1): Decimal('1000.0'),  # bill 4
                    date(2018, 4, 1): Decimal('1500.0')  # bill 11
                },
                increases={
                    date(2017, 8, 1): Decimal('1200.0'),  # bill 3
                    date(2018, 2, 1): Decimal('1500.0'),  # bill 9
                }
            ),
            [self.stmt_cc_one, self.stmt_cc_two, self.stmt_cc_three]
        )
        assert res == [
            (3, Decimal('1030.0'), Decimal('500.00')),
            (4, Decimal('2160.0'), Decimal('200.00')),
            (11, Decimal('11100.0'), Decimal('500.00'))
        ]

    def test_highest_balance_first_with_increases_and_onetimes(self):
        res = calculate_payoffs(
            HighestBalanceFirstMethod(
                Decimal('1000.0'),
                onetimes={
                    date(2017, 10, 1): Decimal('7000.0'),  # bill 5
                    date(2017, 11, 1): Decimal('240.0'),  # bill 6
                },
                increases={
                    date(2017, 8, 1): Decimal('1200.0'),  # bill 3
                    date(2018, 2, 1): Decimal('1500.0'),  # bill 9
                }
            ),
            [self.stmt_cc_one, self.stmt_cc_two, self.stmt_cc_three]
        )
        assert res == [
            (6, Decimal('1060.0'), Decimal('200.00')),
            (6, Decimal('2240.0'), Decimal('200.00')),
            (5, Decimal('10500.0'), Decimal('600.00'))
        ]


class InterestData(object):
    """
    Some example data for interest and payoff calculations.
    """

    citi = [
        {
            'start': date(2016, 12, 24),
            'end': date(2017, 1, 24),
            'apr': Decimal("0.1574"),
            'start_bal': Decimal("5542.78"),
            'bal_subj_to_int': Decimal("5551.84"),
            'transactions': {
                date(2017, 1, 20): -180,
            },
            'interest_amt': Decimal("76.61"),
            'end_balance': Decimal("5439.39"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("130.61"), 19, 11650],
                [191, 3, 6876]
            ]
        },
        {
            'start': date(2017, 1, 25),
            'end': date(2017, 2, 23),
            'apr': Decimal("0.1574"),
            'start_bal': Decimal("5439.39"),
            'bal_subj_to_int': Decimal("5499.95"),
            'transactions': {
                date(2017, 2, 17): -140,
                date(2017, 1, 25): Decimal("58.75")
            },
            'interest_amt': Decimal("71.15"),
            'end_balance': Decimal("5429.29"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("125.15"), 19, 11628],
                [190, 3, 6840]
            ]
        },
        {
            'start': date(2017, 2, 24),
            'end': date(2017, 3, 23),
            'apr': Decimal("0.1599"),
            'start_bal': Decimal("5429.29"),
            'bal_subj_to_int': Decimal("5419.22"),
            'transactions': {
                date(2017, 3, 17): -169,
            },
            'interest_amt': Decimal("66.48"),
            'end_balance': Decimal("5326.77"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("119.48"), 19, 11506],
                [187, 3, 6732]
            ]
        },
        {
            'start': date(2017, 3, 24),
            'end': date(2017, 4, 25),
            'apr': Decimal("0.1599"),
            'start_bal': Decimal("5326.77"),
            'bal_subj_to_int': Decimal("5320.73"),
            'transactions': {
                date(2017, 4, 14): Decimal("-119.48"),
            },
            'interest_amt': Decimal("76.92"),
            'end_balance': Decimal("5284.21"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("128.92"), 19, 11393],
                [186, 3, 6696]
            ]
        },
        {
            'start': date(2017, 4, 26),
            'end': date(2017, 5, 23),
            'apr': Decimal("0.1599"),
            'start_bal': Decimal("5284.21"),
            'bal_subj_to_int': Decimal("5260.20"),
            'transactions': {
                date(2017, 5, 12): Decimal("-128.92"),
            },
            'interest_amt': Decimal("64.53"),
            'end_balance': Decimal("5219.82"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("116.53"), 19, 11259],
                [183, 3, 6588]
            ]
        },
        {
            'start': date(2017, 5, 24),
            'end': date(2017, 6, 23),
            'apr': Decimal("0.1624"),
            'start_bal': Decimal("5219.82"),
            'bal_subj_to_int': Decimal("5192.23"),
            'transactions': {
                date(2017, 6, 9): Decimal("-128.92"),
            },
            'interest_amt': Decimal("71.61"),
            'end_balance': Decimal("5162.51"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("122.61"), 19, 11209],
                [182, 3, 6552]
            ]
        },
        {
            'start': date(2017, 6, 24),
            'end': date(2017, 7, 25),
            'apr': Decimal("0.1624"),
            'start_bal': Decimal("5162.51"),
            'bal_subj_to_int': Decimal("5178.1"),
            'transactions': {
                date(2017, 7, 21): -129,
            },
            'interest_amt': Decimal("73.72"),
            'end_balance': Decimal("5107.23"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("124.72"), 18, 11079],
                [180, 3, 6480]
            ]
        },
        {
            'start': date(2016, 7, 26),
            'end': date(2016, 8, 23),
            'apr': Decimal("0.1549"),
            'start_bal': Decimal("5450.1"),
            'bal_subj_to_int': Decimal("5490.59"),
            'transactions': {
                date(2016, 8, 19): Decimal("-128.64"),
                date(2016, 8, 1): Decimal("13.89"),
                date(2016, 8, 3): (Decimal("18.89") + Decimal("7.39"))
            },
            'interest_amt': Decimal("67.58"),
            'end_balance': Decimal("5429.21"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("121.58"), 19, 11529],
                [190, 3, 6840]
            ]
        },
        {
            'start': date(2016, 8, 24),
            'end': date(2016, 9, 23),
            'apr': Decimal("0.1549"),
            'start_bal': Decimal("5429.21"),
            'bal_subj_to_int': Decimal("5469.66"),
            'transactions': {
                date(2016, 9, 16): -200,
                date(2016, 8, 27): Decimal("59.8"),
                date(2016, 9, 9): Decimal("6.41")
            },
            'interest_amt': Decimal("71.96"),
            'end_balance': Decimal("5367.38"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("124.96"), 19, 11380],
                [187, 3, 6732]
            ]
        },
        {
            'start': date(2016, 9, 24),
            'end': date(2016, 10, 25),
            'apr': Decimal("0.1549"),
            'start_bal': Decimal("5367.38"),
            'bal_subj_to_int': Decimal("5423.19"),
            'transactions': {
                date(2016, 10, 14): -180,
                date(2016, 9, 27): Decimal("21.9"),
                date(2016, 10, 6): Decimal("76.93"),
                date(2016, 10, 7): Decimal("33.16")
            },
            'interest_amt': Decimal("73.65"),
            'end_balance': Decimal("5393.02"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("126.65"), 19, 11440],
                [188, 3, 6768]
            ]
        },
        {
            'start': date(2016, 10, 26),
            'end': date(2016, 11, 23),
            'apr': Decimal("0.1549"),
            'start_bal': Decimal("5393.02"),
            'bal_subj_to_int': Decimal("5422.18"),
            'transactions': {
                date(2016, 11, 10): -180,
                date(2016, 10, 26): Decimal("17.52"),
                date(2016, 11, 4): Decimal("85.57"),
                date(2016, 11, 18): 35
            },
            'interest_amt': Decimal("66.73"),
            'end_balance': Decimal("5417.84"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("120.73"), 19, 11500],
                [189, 3, 6804]
            ]
        },
        {
            'start': date(2016, 11, 24),
            'end': date(2016, 12, 23),
            'apr': Decimal("0.1574"),
            'start_bal': Decimal("5417.84"),
            'bal_subj_to_int': Decimal("5464.87"),
            'transactions': {
                date(2016, 12, 9): -160,
                date(2016, 11, 24): Decimal("43.31"),
                date(2016, 12, 11): 44,
                date(2016, 12, 14): Decimal("82.68"),
                date(2016, 12, 22): Decimal("44.26")
            },
            'interest_amt': Decimal("70.69"),
            'end_balance': Decimal("5542.78"),
            'payoffs': [
                # amt, time, total_paid
                [Decimal("168.47"), 19, 11840],
                [194, 3, 6984]
            ]
        }
    ]

    amex = [
        {
            'start': date(2017, 7, 28) - timedelta(days=30),
            'end': date(2017, 7, 28),
            'apr': Decimal("0.1824"),
            'start_bal': Decimal("1856.23"),
            'bal_subj_to_int': Decimal("1821.35"),
            'transactions': {
                date(2017, 7, 21): Decimal("-189.0"),
            },
            'interest_amt': Decimal("28.23"),
            'end_balance': Decimal("1695.46"),
            'payoffs': [
                # amt, time, total_paid
                [45, 7, 2924],
                [62, 3, 2216]
            ]
        },
        {
            'start': date(2017, 4, 27) - timedelta(days=29),
            'end': date(2017, 4, 27),
            'apr': Decimal("0.1799"),
            'start_bal': Decimal("1917.31"),
            'bal_subj_to_int': Decimal("1908.13"),
            'transactions': {
                date(2017, 4, 14): Decimal("-49.0"),
            },
            'interest_amt': Decimal("28.22"),
            'end_balance': Decimal("1896.53"),
            'payoffs': [
                # amt, time, total_paid
                [47, 8, 3403],
                [69, 3, 2471]
            ]
        },
        {
            'start': date(2017, 2, 24) - timedelta(days=27),
            'end': date(2017, 2, 24),
            'apr': Decimal("0.1774"),
            'start_bal': Decimal("1977.06"),
            'bal_subj_to_int': Decimal("1969.76"),
            'transactions': {
                date(2017, 2, 17): Decimal("-71.0"),
            },
            'interest_amt': Decimal("26.8"),
            'end_balance': Decimal("1932.86"),
            'payoffs': [
                # amt, time, total_paid
                [46, 8, 3466],
                [70, 3, 2509]
            ]
        },
        {
            'start': date(2017, 6, 27) - timedelta(days=29),
            'end': date(2017, 6, 27),
            'apr': Decimal("0.1824"),
            'start_bal': Decimal("1878.33"),
            'bal_subj_to_int': Decimal("1860.2"),
            'transactions': {
                date(2017, 6, 9): Decimal("-50.0"),
            },
            'interest_amt': Decimal("27.9"),
            'end_balance': Decimal("1856.23"),
            'payoffs': [
                # amt, time, total_paid
                [46, 8, 3334],
                [67, 3, 2427]
            ]
        },
        {
            'start': date(2017, 3, 28) - timedelta(days=31),
            'end': date(2017, 3, 28),
            'apr': Decimal("0.1799"),
            'start_bal': Decimal("1932.86"),
            'bal_subj_to_int': Decimal("1930.4"),
            'transactions': {
                date(2017, 3, 17): Decimal("-46.0"),
            },
            'interest_amt': Decimal("30.45"),
            'end_balance': Decimal("1917.31"),
            'payoffs': [
                # amt, time, total_paid
                [49, 8, 3452],
                [69, 3, 2498]
            ]
        },
        {
            'start': date(2017, 5, 28) - timedelta(days=30),
            'end': date(2017, 5, 28),
            'apr': Decimal("0.1799"),
            'start_bal': Decimal("1896.53"),
            'bal_subj_to_int': Decimal("1884.74"),
            'transactions': {
                date(2017, 5, 12): Decimal("-47.0"),
            },
            'interest_amt': Decimal("28.8"),
            'end_balance': Decimal("1878.33"),
            'payoffs': [
                # amt, time, total_paid
                [47, 8, 3357],
                [68, 3, 2447]
            ]
        }
    ]

    discover = [
        {
            'start': date(2016, 12, 19),
            'end': date(2017, 1, 18),
            'apr': Decimal("0.1949"),
            'start_bal': Decimal("7160.59"),
            'bal_subj_to_int': Decimal("7149.14"),
            'transactions': {
                date(2017, 1, 6): -150,
                date(2016, 12, 29): Decimal("9.98"),
                date(2017, 1, 16): Decimal("6.86"),
                date(2017, 1, 3): 10,
                date(2017, 1, 17): Decimal("67.5")
            },
            'interest_amt': Decimal("118.53"),
            'end_balance': Decimal("7212.01"),
            'payoffs': [
                # amt, time, total_paid
                [145, 28, 26912],
                [266, 3, 9582]
            ]
        },
        {
            'start': date(2017, 1, 19),
            'end': date(2017, 2, 18),
            'apr': Decimal("0.1949"),
            'start_bal': Decimal("7236.56"),
            'bal_subj_to_int': Decimal("7212.01"),
            'transactions': {
                date(2017, 2, 3): Decimal("-184.95"),
                date(2017, 2, 7): Decimal("21.12"),
                date(2017, 1, 17): Decimal("53.7")
            },
            'interest_amt': Decimal("119.8"),
            'end_balance': Decimal("7221.68"),
            'payoffs': [
                # amt, time, total_paid
                [145, 28, 26952],
                [267, 3, 9595]
            ]
        },
        {
            'start': date(2017, 2, 19),
            'end': date(2017, 3, 18),
            'apr': Decimal("0.1949"),
            'start_bal': Decimal("7190.79"),
            'bal_subj_to_int': Decimal("7221.68"),
            'transactions': {
                date(2017, 3, 3): Decimal("-145.0"),
            },
            'interest_amt': Decimal("107.52"),
            'end_balance': Decimal("7184.2"),
            'payoffs': [
                # amt, time, total_paid
                [144, 28, 26800],
                [265, 3, 9588]
            ]
        },
        {
            'start': date(2017, 3, 19),
            'end': date(2017, 4, 18),
            'apr': Decimal("0.1974"),
            'start_bal': Decimal("7154.09"),
            'bal_subj_to_int': Decimal("7184.2"),
            'transactions': {
                date(2017, 3, 31): Decimal("-144.0"),
            },
            'interest_amt': Decimal("119.94"),
            'end_balance': Decimal("7160.14"),
            'payoffs': [
                # amt, time, total_paid
                [144, 28, 27291],
                [265, 3, 9545]
            ]
        },
        {
            'start': date(2017, 4, 19),
            'end': date(2017, 5, 18),
            'apr': Decimal("0.1974"),
            'start_bal': Decimal("7182.91"),
            'bal_subj_to_int': Decimal("7160.14"),
            'transactions': {
                date(2017, 5, 12): Decimal("-144.0"),
            },
            'interest_amt': Decimal("116.54"),
            'end_balance': Decimal("7132.68"),
            'payoffs': [
                # amt, time, total_paid
                [143, 28, 27144],
                [264, 3, 9509]
            ]
        },
        {
            'start': date(2017, 5, 19),
            'end': date(2017, 6, 18),
            'apr': Decimal("0.1974"),
            'start_bal': Decimal("7144.28"),
            'bal_subj_to_int': Decimal("7132.68"),
            'transactions': {
                date(2017, 6, 9): Decimal("-144.0"),
            },
            'interest_amt': Decimal("119.77"),
            'end_balance': Decimal("7108.45"),
            'payoffs': [
                # amt, time, total_paid
                [143, 28, 26991],
                [263, 3, 9476]
            ]
        },
        {
            'start': date(2017, 6, 19),
            'end': date(2017, 7, 18),
            'apr': Decimal("0.1999"),
            'start_bal': Decimal("7107.42"),
            'bal_subj_to_int': Decimal("7108.45"),
            'transactions': {
                date(2017, 7, 7): Decimal("-144.0"),
            },
            'interest_amt': Decimal("116.78"),
            'end_balance': Decimal("7081.23"),
            'payoffs': [
                # amt, time, total_paid
                [142, 28, 27398],
                [263, 3, 9472]
            ]
        }
    ]


@pytest.mark.skipif(PY34PLUS is False,
                    reason='py3.4+ only due to Decimal rounding')
class TestDataAmEx(object):

    def test_calculate(self, data):
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentAmEx()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = deepcopy(data['start'])
        res = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions']
        )
        assert res.start_date == data['start']
        assert res.end_date == data['end']
        assert pctdiff(
            round(res.principal, 2), data['end_balance']) < Decimal('0.01')
        assert pctdiff(
            round(res.interest, 2), data['interest_amt']) < Decimal('0.01')

    def test_calculate_min_payment(self, data):
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentAmEx()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = deepcopy(data['start'])
        res = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions'],
            end_balance=data['end_balance'],
            interest_amt=data['interest_amt']
        )
        assert res.start_date == data['start']
        assert res.end_date == data['end']
        assert int(res.minimum_payment) == data['payoffs'][0][0]

    def test_calculate_payoff_min(self, data):
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentAmEx()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = deepcopy(data['start'])
        stmt = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions'],
            end_balance=data['end_balance'],
            interest_amt=data['interest_amt']
        )
        res = calculate_payoffs(
            MinPaymentMethod(),
            [stmt]
        )
        if data['end'] == date(2017, 6, 27):
            # difference based on our estimated billing periods
            data['payoffs'][0][1] = 7
        assert stmt.start_date == data['start']
        assert stmt.end_date == data['end']
        assert int(round(res[0][0]/12)) == data['payoffs'][0][1]
        assert pctdiff(res[0][1], data['payoffs'][0][2]) < Decimal('0.03')

    def test_calculate_payoff_recommended(self, data):
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentAmEx()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = deepcopy(data['start'])
        stmt = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions'],
            end_balance=data['end_balance'],
            interest_amt=data['interest_amt']
        )
        res = calculate_payoffs(
            FixedPaymentMethod(data['payoffs'][1][0]),
            [stmt]
        )
        assert stmt.start_date == data['start']
        assert stmt.end_date == data['end']
        assert int(round(res[0][0]/12)) == data['payoffs'][1][1]
        assert pctdiff(res[0][1], data['payoffs'][1][2]) < Decimal('0.03')


@pytest.mark.skipif(PY34PLUS is False,
                    reason='py3.4+ only due to Decimal rounding')
class TestDataCiti(object):

    def test_calculate(self, data):
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentCiti()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = data['start']
        res = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions']
        )
        assert res.start_date == data['start']
        assert res.end_date == data['end']
        assert pctdiff(
            round(res.principal, 2), data['end_balance']) < Decimal('0.01')
        assert pctdiff(
            round(res.interest, 2), data['interest_amt']) < Decimal('0.01')

    def test_calculate_min_payment(self, data):
        if data['end'] == date(2016, 12, 23):
            pytest.skip("Strange minimum payment; outliar.")
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentCiti()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = data['start']
        res = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions'],
            end_balance=data['end_balance'],
            interest_amt=data['interest_amt']
        )
        assert res.start_date == data['start']
        assert res.end_date == data['end']
        assert pctdiff(
            res.minimum_payment, data['payoffs'][0][0]
        ) < Decimal('0.01')

    def test_calculate_payoff_min(self, data):
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentCiti()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = data['start']
        stmt = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions'],
            end_balance=data['end_balance'],
            interest_amt=data['interest_amt']
        )
        res = calculate_payoffs(
            MinPaymentMethod(),
            [stmt]
        )
        assert stmt.start_date == data['start']
        assert stmt.end_date == data['end']
        if data['end'] == date(2017, 7, 25):
            pytest.skip('Strange data.')
        assert ceil(res[0][0]/12) == data['payoffs'][0][1]
        assert pctdiff(res[0][1], data['payoffs'][0][2]) < Decimal('0.03')

    def test_calculate_payoff_recommended(self, data):
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentCiti()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = data['start']
        stmt = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions'],
            end_balance=data['end_balance'],
            interest_amt=data['interest_amt']
        )
        res = calculate_payoffs(
            FixedPaymentMethod(data['payoffs'][1][0]),
            [stmt]
        )
        assert stmt.start_date == data['start']
        assert stmt.end_date == data['end']
        assert int(round(res[0][0]/12)) == data['payoffs'][1][1]
        assert pctdiff(res[0][1], data['payoffs'][1][2]) < Decimal('0.03')


@pytest.mark.skipif(PY34PLUS is False,
                    reason='py3.4+ only due to Decimal rounding')
class TestDataDiscover(object):

    def test_calculate(self, data):
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentDiscover()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = data['start']
        res = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions']
        )
        assert res.start_date == data['start']
        assert res.end_date == data['end']
        assert pctdiff(
            round(res.principal, 2), data['end_balance']) < Decimal('0.01')
        assert pctdiff(
            round(res.interest, 2), data['interest_amt']) < Decimal('0.01')

    def test_calculate_min_payment(self, data):
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentDiscover()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = data['start']
        res = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions'],
            end_balance=data['end_balance'],
            interest_amt=data['interest_amt']
        )
        assert res.start_date == data['start']
        assert res.end_date == data['end']
        assert pctdiff(
            res.minimum_payment, data['payoffs'][0][0]
        ) < Decimal('0.01')

    def test_calculate_payoff_min(self, data):
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentDiscover()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = data['start']
        stmt = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions'],
            end_balance=data['end_balance'],
            interest_amt=data['interest_amt']
        )
        res = calculate_payoffs(
            MinPaymentMethod(),
            [stmt]
        )
        # fix discover calculation
        data['payoffs'][0][1] -= 1
        assert stmt.start_date == data['start']
        assert stmt.end_date == data['end']
        assert int(round(res[0][0]/12)) == data['payoffs'][0][1]
        assert pctdiff(res[0][1], data['payoffs'][0][2]) < Decimal('0.04')

    def test_calculate_payoff_recommended(self, data):
        icls = AdbCompoundedDaily(data['apr'])
        mpc = MinPaymentDiscover()
        bp = _BillingPeriod(data['end'])
        bp._end_date = deepcopy(data['end'])
        bp._start_date = data['start']
        stmt = CCStatement(
            icls,
            data['start_bal'],
            mpc,
            bp,
            transactions=data['transactions'],
            end_balance=data['end_balance'],
            interest_amt=data['interest_amt']
        )
        res = calculate_payoffs(
            FixedPaymentMethod(data['payoffs'][1][0]),
            [stmt]
        )
        assert stmt.start_date == data['start']
        assert stmt.end_date == data['end']
        assert int(round(res[0][0]/12)) == data['payoffs'][1][1]
        assert pctdiff(res[0][1], data['payoffs'][1][2]) < Decimal('0.03')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
import os
import logging
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
import socket
import json
from time import time
import warnings
import multiprocessing

from retrying import retry

import biweeklybudget.settings

# Python 3.14+ defaults to 'forkserver' multiprocessing start method,
# but pytest-flask's LiveServer requires 'fork' for pickling
try:
    multiprocessing.set_start_method('fork')
except RuntimeError:
    # Start method can only be set once
    pass
from biweeklybudget.tests.fixtures.sampledata import SampleDataLoader
from biweeklybudget.tests.sqlhelpers import restore_mysqldump, do_mysqldump
from biweeklybudget.tests.selenium_helpers import (
    set_browser_for_fullpage_screenshot
)

try:
    # Note that in pytest 3.6.0 thanks to issue #3487, anything called
    # "pytest_" in this file is attempted to be loaded as a plugin, and then
    # causes the test run to fail.
    import pytest_selenium.pytest_selenium as ptselenium
    from selenium.webdriver.support.event_firing_webdriver import \
        EventFiringWebDriver
    from selenium.webdriver.chrome.webdriver import WebDriver as ChromeWD
    from selenium.webdriver import ChromeOptions
    HAVE_PYTEST_SELENIUM = True
except ImportError:
    HAVE_PYTEST_SELENIUM = False

import biweeklybudget.db  # noqa
import biweeklybudget.models.base  # noqa
from biweeklybudget.db_event_handlers import init_event_listeners  # noqa
from biweeklybudget.tests.unit.test_interest import InterestData  # noqa
from biweeklybudget.tests.migrations.alembic_helpers import (
    uri_for_db, empty_db_by_uri  # noqa
)

# Monkey-patch datatables package for SQLAlchemy 2.0 compatibility
# The package uses query.join('relationship_name') which no longer works in
# SQLAlchemy 2.0 - it must pass the actual relationship attribute instead
import datatables
from collections import namedtuple

_original_datatable_init = datatables.DataTable.__init__


def _patched_datatable_init(self, params, model, query, columns):
    self.params = params
    self.model = model
    self.query = query
    self.data = {}
    self.columns = []
    self.columns_dict = {}
    self.search_func = lambda qs, s: qs

    DataColumn = namedtuple("DataColumn", ("name", "model_name", "filter"))

    for col in columns:
        name, model_name, filter_func = None, None, None

        if isinstance(col, datatables.DataColumn):
            self.columns.append(col)
            continue
        elif isinstance(col, tuple):
            if len(col) == 3:
                name, model_name, filter_func = col
            elif len(col) == 2:
                if callable(col[1]):
                    name, filter_func = col
                    model_name = name
                else:
                    name, model_name = col
            else:
                raise ValueError("Columns must be a tuple of 2 to 3 elements")
        else:
            name, model_name = col, col

        d = DataColumn(name=name, model_name=model_name, filter=filter_func)
        self.columns.append(d)
        self.columns_dict[d.name] = d

    # SQLAlchemy 2.0 fix: use getattr to get the actual relationship attribute
    # instead of passing a string to query.join()
    for column in (col for col in self.columns if "." in col.model_name):
        relationship_name = column.model_name.split(".")[0]
        self.query = self.query.join(getattr(self.model, relationship_name))


datatables.DataTable.__init__ = _patched_datatable_init

LIVESERVER_LOG_PATH = os.environ.get('BIWEEKLYBUDGET_LOG_FILE')

_DB_ENGINE = None

logger = logging.getLogger(__name__)

# suppress webdriver DEBUG logging
selenium_log = logging.getLogger("selenium")
selenium_log.setLevel(logging.INFO)
selenium_log.propagate = True

class_refresh_db_durations = []


def get_db_engine():
    global _DB_ENGINE
    if _DB_ENGINE is None:
        connstr = os.environ.get('DB_CONNSTRING', None)
        if connstr is None:
            connstr = 'mysql+pymysql://budgetTester:jew8fu0ue@127.0.0.1:3306/' \
                      'budgettest?charset=utf8mb4'
            os.environ['DB_CONNSTRING'] = connstr
        biweeklybudget.settings.DB_CONNSTRING = connstr
        _DB_ENGINE = create_engine(
            connstr, echo=False,
            connect_args={
                'sql_mode': 'STRICT_ALL_TABLES,NO_ZERO_DATE,'
                            'NO_ZERO_IN_DATE,ERROR_FOR_DIVISION_BY_ZERO,'
                            'NO_AUTO_CREATE_USER'
            },
            pool_size=10, pool_timeout=120
        )
    return _DB_ENGINE


@pytest.fixture
def alembic_root():
    return os.path.join(
        os.path.abspath(os.environ['TOXINIDIR']), 'biweeklybudget', 'alembic'
    )


@pytest.fixture
def uri_left():
    uri = uri_for_db(os.environ['MYSQL_DBNAME_LEFT'])
    empty_db_by_uri(uri)
    return uri


@pytest.fixture
def uri_right():
    uri = uri_for_db(os.environ['MYSQL_DBNAME_RIGHT'])
    empty_db_by_uri(uri)
    return uri


@pytest.fixture(scope='session')
def dump_file_path(tmpdir_factory):
    """
    Return the directory to use for the SQL dump files
    """
    return tmpdir_factory.mktemp('sqldump')


@pytest.fixture(scope="session")
def refreshdb(dump_file_path):
    """
    Refresh/Load DB data before tests; also exec mysqldump to write a
    SQL dump file for faster refreshes during test runs.
    """
    if 'NO_REFRESH_DB' not in os.environ:
        # setup the connection
        conn = get_db_engine().connect()
        logger.info('Refreshing DB (session-scoped)')
        # clean the database
        biweeklybudget.models.base.Base.metadata.drop_all(get_db_engine())
        with warnings.catch_warnings():
            # Ignore warning from MariaDB MDEV-17544 when creating tables;
            # SQLAlchemy 1.3.13 names the primary keys, but MariaDB ignores
            # the names; MariaDB >= 10.4.7 now throws a warning for this.
            warnings.filterwarnings(
                'ignore',
                message=r'^\(1280, "Name.*ignored for PRIMARY key\."\)$',
                category=Warning
            )
            biweeklybudget.models.base.Base.metadata.create_all(get_db_engine())
        # load the sample data
        data_sess = scoped_session(
            sessionmaker(autocommit=False, autoflush=False, bind=conn)
        )
        SampleDataLoader(data_sess).load()
        data_sess.flush()
        data_sess.commit()
        data_sess.close()
        # close connection
        conn.close()
    else:
        logger.info('Skipping session-scoped DB refresh')
    # write the dump files
    do_mysqldump(dump_file_path, get_db_engine())
    do_mysqldump(dump_file_path, get_db_engine(), with_data=False)
    yield


@pytest.fixture(scope="class")
def class_refresh_db(dump_file_path):
    """
    This fixture rolls the DB back to the previous state when the class is
    finished; to be used on classes that alter data.

    Use like:

        @pytest.mark.usefixtures('class_refresh_db', 'testdb')
        class MyClass(AcceptanceHelper):
    """
    global class_refresh_db_durations
    yield
    if 'NO_CLASS_REFRESH_DB' in os.environ:
        return
    logger.info('Refreshing DB (class-scoped)')
    s = time()
    restore_mysqldump(dump_file_path, get_db_engine())
    class_refresh_db_durations.append(time() - s)


@pytest.fixture
def testdb():
    """
    DB fixture to be used in tests
    """
    # setup the connection
    engine = get_db_engine()
    conn = engine.connect()
    sess = sessionmaker(autocommit=False, bind=conn)()
    init_event_listeners(sess, engine)
    # yield the session
    yield sess
    sess.close()
    conn.close()


@pytest.fixture(scope="session")
def testflask():
    """
    This is a version of pytest-flask's live_server fixture, modified for
    session use.

    When BIWEEKLYBUDGET_TEST_BASE_URL is set (e.g., for Docker tests),
    we don't start a live server - tests run against the external URL instead.
    """
    if 'BIWEEKLYBUDGET_TEST_BASE_URL' in os.environ:
        # When running against an external URL (e.g., Docker container),
        # we don't need to start a local server
        yield None
    else:
        # Bind to an open port
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(('', 0))
        port = s.getsockname()[1]
        s.close()
        from pytest_flask.fixtures import LiveServer  # noqa
        from biweeklybudget.flaskapp.app import app  # noqa
        server = LiveServer(app, 'localhost', port, wait=5)
        server.start()
        yield server
        server.stop()


@pytest.fixture(scope="session")
def base_url(testflask):
    """
    Simple fixture to return ``testflask`` base URL
    """
    url = os.environ.get('BIWEEKLYBUDGET_TEST_BASE_URL', None)
    if url is not None:
        return url
    return testflask.url()


def retry_if_conn_error(ex):
    try:
        return isinstance(ex, ConnectionError)
    except NameError:
        return isinstance(ex, IOError)


@retry(
    stop_max_attempt_number=3, wait_fixed=3000,
    retry_on_exception=retry_if_conn_error
)
def get_driver_for_class(driver_class, driver_kwargs):
    """
    Wrapper around the selenium ``driver()`` fixture's
    ``driver_class(**driver_kwargs)`` call that retries up to 3 times, 3
    seconds apart, if a ConnectionError is raised.
    """
    return driver_class(**driver_kwargs)


@pytest.fixture
def driver(request, driver_class, driver_kwargs):
    """
    Returns a WebDriver instance based on options and capabilities

    This is copied from pytest-selenium 1.11.4, but modified to retry getting
    the driver up to 3 times to cope with intermittent connection resets in
    TravisCI. We ripped the original ``driver = driver_class(**driver_kwargs)``
    out and replaced it with the ``get_driver_for_class()`` function, which
    is wrapped in the retrying package's ``@retry`` decorator.

    Updated for Selenium 4.x which removed desired_capabilities parameter.
    """
    kwargs = driver_kwargs.copy()
    # Selenium 4.x: desired_capabilities parameter removed, use options instead
    if 'desired_capabilities' in kwargs:
        del kwargs['desired_capabilities']

    # Set logging preferences through ChromeOptions for Selenium 4.x
    if 'options' not in kwargs:
        kwargs['options'] = ChromeOptions()
    kwargs['options'].set_capability('goog:loggingPrefs', {
        'browser': 'ALL',
        'driver': 'ALL'
    })

    driver = get_driver_for_class(driver_class, kwargs)
    event_listener = request.config.getoption('event_listener')
    if event_listener is not None:
        # Import the specified event listener and wrap the driver instance
        mod_name, class_name = event_listener.rsplit('.', 1)
        mod = __import__(mod_name, fromlist=[class_name])
        event_listener = getattr(mod, class_name)
        if not isinstance(driver, EventFiringWebDriver):
            driver = EventFiringWebDriver(driver, event_listener())

    request.node._driver = driver
    yield driver
    driver.quit()


@pytest.fixture
def selenium(driver):
    """
    Per pytest-selenium docs, use this to override the selenium fixture to
    provide global common setup.
    """
    driver.set_window_size(1920, 1080)
    driver.implicitly_wait(2)
    # from http://stackoverflow.com/a/13853684/211734
    driver.set_script_timeout(30)
    # from http://stackoverflow.com/a/17536547/211734
    driver.set_page_load_timeout(30)
    yield driver


def _gather_screenshot(item, report, driver, summary, extra):
    """
    Redefine pytest-selenium's _gather_screenshot so that we can get full-page
    screenshots. This implementation is copied from pytest-selenium 1.11.4,
    but calls :py:func:`~.selenium_helpers.set_browser_for_fullpage_screenshot`
    before calling ``driver.get_screenshot_as_base64()``.
    """
    try:
        set_browser_for_fullpage_screenshot(driver)
        screenshot = driver.get_screenshot_as_base64()
    except Exception as e:
        summary.append('WARNING: Failed to gather screenshot: {0}'.format(e))
        return
    pytest_html = item.config.pluginmanager.getplugin('html')
    if pytest_html is not None:
        # add screenshot to the html report
        extra.append(pytest_html.extras.image(screenshot, 'Screenshot'))


def _gather_logs(item, report, driver, summary, extra):
    """Redefine here to fix a bug"""
    pytest_html = item.config.pluginmanager.getplugin("html")
    try:
        # this doesn't work with chromedriver
        # types = driver.log_types
        types = ['browser', 'driver', 'server']
    except Exception as e:
        # note that some drivers may not implement log types
        summary.append("WARNING: Failed to gather log types: {0}".format(e))
        return
    for name in types:
        try:
            log = driver.get_log(name)
        except Exception as e:
            summary.append(
                "WARNING: Failed to gather {0} log: {1}".format(name, e)
            )
            return
        if pytest_html is not None:
            extra.append(
                pytest_html.extras.text(
                    ptselenium.format_log(log), "%s Log" % name.title()
                )
            )


# redefine _gather_screenshot to use our implementation
if HAVE_PYTEST_SELENIUM:
    ptselenium._gather_screenshot = _gather_screenshot
    ptselenium._gather_logs = _gather_logs


@pytest.fixture
def chrome_options(chrome_options: 'ChromeOptions') -> 'ChromeOptions':
    chrome_options.add_argument('headless')
    chrome_options.add_argument('window-size=1920x1080')
    # Add stability options for CI environments
    chrome_options.add_argument('--disable-dev-shm-usage')
    chrome_options.add_argument('--no-sandbox')
    chrome_options.add_argument('--disable-gpu')
    chrome_options.add_argument('--disable-software-rasterizer')
    return chrome_options


def pytest_addoption(parser):
    group = parser.getgroup("terminal reporting", "reporting", after="general")
    group.addoption('--durations-file',
                    action="store", type=str, default=None, metavar="P",
                    help="write all durations as JSON to P")


def pytest_terminal_summary(terminalreporter):
    """Write all test durations to results/durations.json"""
    fpath = terminalreporter.config.option.durations_file
    if fpath is None:
        return
    tr = terminalreporter
    dlist = []
    for replist in tr.stats.values():
        for rep in replist:
            if hasattr(rep, 'duration'):
                dlist.append(rep)
    if not dlist:
        return
    dlist.sort(key=lambda x: x.duration)
    dlist.reverse()
    tr.write_sep("-", "wrote all test durations to: %s" % fpath)

    result = []
    for rep in dlist:
        nodeid = rep.nodeid.replace("::()::", "::")
        result.append([nodeid, rep.when, rep.duration])
    with open(fpath, 'w') as fh:
        fh.write(json.dumps({
            'requests': result,
            'class_refresh_db': class_refresh_db_durations
        }))


# next section from:
# https://docs.pytest.org/en/latest/example/simple.html#\
# incremental-testing-test-steps


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    if "incremental" in item.keywords:
        if call.excinfo is not None:
            parent = item.parent
            parent._previousfailed = item
    outcome = yield
    if (
        hasattr(item.session, 'liveserver_log') and
        item.nodeid in item.session.liveserver_log
    ):
        report = outcome.get_result()
        extra = getattr(report, 'extra', [])
        pos = item.session.liveserver_log[item.nodeid]
        pytest_html = item.config.pluginmanager.getplugin('html')
        if pytest_html is not None:
            fh = open(LIVESERVER_LOG_PATH, 'r')
            fh.seek(pos['start'])
            if 'end' in pos:
                log = fh.read(pos['end'] - pos['start'])
            else:
                log = fh.read()
            extra.append(pytest_html.extras.text(log, 'LiveServer'))
            msg = 'NOTE: testflask LiveServer logging captured via ' \
                  'pytest-html to: %s %s' % (LIVESERVER_LOG_PATH, pos)
            for item in list(report.sections):
                if item[0] == 'testflask':
                    report.sections.remove(item)
            report.sections.append(('testflask', msg))
            report.extra = extra
        else:
            report.sections.append((
                'testflask',
                'ERROR - pytest-html not present, cannot save server log!\n'
            ))


def getfilesize(fpath):
    f = open(fpath, 'r')
    f.seek(0, 2)
    val = f.tell()
    f.close()
    return val


def pytest_runtest_setup(item):
    if "incremental" in item.keywords:
        previousfailed = getattr(item.parent, "_previousfailed", None)
        if previousfailed is not None:
            pytest.xfail("previous test failed (%s)" % previousfailed.name)
    if hasattr(item, 'fixturenames') and 'testflask' in item.fixturenames:
        if not hasattr(item.session, 'liveserver_log'):
            setattr(item.session, 'liveserver_log', {})
        item.session.liveserver_log[item.nodeid] = {
            'start': getfilesize(LIVESERVER_LOG_PATH)
        }


def pytest_runtest_teardown(item):
    if (
        hasattr(item.session, 'liveserver_log') and
        item.nodeid in item.session.liveserver_log
    ):
        item.session.liveserver_log[item.nodeid][
            'end'] = getfilesize(LIVESERVER_LOG_PATH)


"""
Begin generated/parametrized tests for interest calculation.

I REALLY wish pytest still supported yield tests, this is sooooo messy!
"""


def pytest_generate_tests(metafunc):
    if (
        metafunc.function.__name__.startswith('test_calculate') and
        metafunc.module.__name__ == 'biweeklybudget.tests.unit.test_interest'
    ):
        if metafunc.cls.__name__ == 'TestDataAmEx':
            param_for_adbdaily_calc(metafunc, InterestData.amex)
        if metafunc.cls.__name__ == 'TestDataCiti':
            param_for_adbdaily_calc(metafunc, InterestData.citi)
        if metafunc.cls.__name__ == 'TestDataDiscover':
            param_for_adbdaily_calc(metafunc, InterestData.discover)


def param_for_adbdaily_calc(metafunc, s):
    dates = [d['start'].strftime('%Y-%m-%d') for d in s]
    metafunc.parametrize(
        'data',
        s,
        ids=dates
    )



"""base

Revision ID: 7477ff15e8c4
Revises:
Create Date: 2017-04-19 22:19:34.382425

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '7477ff15e8c4'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from sqlalchemy import Column, String
from sqlalchemy_utc import UtcDateTime
from sqlalchemy.orm import relationship

from biweeklybudget.models.base import Base, ModelAsDict

logger = logging.getLogger(__name__)


class PlaidItem(Base, ModelAsDict):

    __tablename__ = 'plaid_items'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key - plaid Item ID
    item_id = Column(String(70), primary_key=True)

    #: Plaid item access token
    access_token = Column(String(70))

    #: institution name
    institution_name = Column(String(100))

    #: institution ID
    institution_id = Column(String(50))

    #: When this item was last updated
    last_updated = Column(UtcDateTime)

    #: Relationship to all :py:class:`~.PlaidAccount` for this Item
    all_accounts = relationship(
        'PlaidAccount', order_by='PlaidAccount.account_id'
    )

    def __repr__(self):
        return "<PlaidItem(item_id=%s, institution_name='%s')>" % (
            self.item_id, self.institution_name
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from sqlalchemy import Column, Integer, Numeric, ForeignKey
from sqlalchemy_utc import UtcDateTime
from sqlalchemy.orm import relationship
from biweeklybudget.models.base import Base, ModelAsDict
from biweeklybudget.utils import dtnow


class AccountBalance(Base, ModelAsDict):

    __tablename__ = 'account_balances'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: ID of the account this balance is for
    account_id = Column(Integer, ForeignKey('accounts.id'), nullable=False)

    #: Relationship to :py:class:`~.Account` this balance is for
    account = relationship(
        "Account", backref="all_balances"
    )

    #: Ledger balance, or investment account value, or credit card balance
    ledger = Column(Numeric(precision=10, scale=4))

    #: as-of date for the ledger balance
    ledger_date = Column(UtcDateTime)

    #: Available balance
    avail = Column(Numeric(precision=10, scale=4))

    #: as-of date for the available balance
    avail_date = Column(UtcDateTime)

    #: overall balance as of DateTime
    overall_date = Column(UtcDateTime, default=dtnow())

    def __repr__(self):
        return "<AccountBalance(id=%d)>" % (
            self.id
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from datetime import datetime
import os
import argparse
import logging
from io import BytesIO
from pytz import UTC

from ofxparse import OfxParser
from sqlalchemy.exc import InvalidRequestError, IntegrityError

from biweeklybudget.cliutils import set_log_debug, set_log_info
from biweeklybudget.ofxapi import apiclient
from biweeklybudget.ofxapi.exceptions import DuplicateFileException

logger = logging.getLogger(__name__)


class OfxBackfiller(object):
    """
    Class to backfill OFX in database from files on disk.
    """

    def __init__(self, client, savedir):
        """
        Initialize the OFX Backfiller.

        :param client: API client
        :type client: Instance of :py:class:`~.OfxApiLocal` or
          :py:class:`~.OfxApiRemote`
        :param savedir: directory/path to save statements in
        :type savedir: str
        """
        logger.info('Initializing OfxBackfiller with savedir=%s', savedir)
        self.savedir = savedir
        self._client = client

    def run(self):
        """
        Main entry point - run the backfill.
        """
        logger.debug('Checking for Accounts with statement directories')
        accounts = self._client.get_accounts()
        for acctname in sorted(accounts.keys()):
            p = os.path.join(self.savedir, acctname)
            data = accounts[acctname]
            if not os.path.isdir(p):
                logger.info('No statement directory for Account %d (%s)',
                            data['id'], p)
                continue
            logger.debug('Found directory %s for Account %d', p, data['id'])
            self._do_account_dir(data['id'], p)

    def _do_account_dir(self, acct_id, path):
        """
        Handle all OFX statements in a per-account directory.

        :param acct_id: account database ID
        :type acct_id: int
        :param path: absolute path to per-account directory
        :type path: str
        """
        logger.debug('Doing account %d directory (%s)', acct_id, path)
        files = {}
        for f in os.listdir(path):
            p = os.path.join(path, f)
            if not os.path.isfile(p):
                continue
            extension = p.split('.')[-1].lower()
            if extension not in ['ofx', 'qfx']:
                continue
            files[p] = os.path.getmtime(p)
        logger.debug('Found %d files for account %d', len(files), acct_id)
        # run through the files, oldest to newest
        success = 0
        already = 0
        for p in sorted(files, key=files.get):
            try:
                self._do_one_file(acct_id, p)
                success += 1
            except DuplicateFileException:
                already += 1
                logger.warning('OFX is already parsed for account; skipping')
            except (InvalidRequestError, IntegrityError, TypeError):
                raise
            except Exception:
                logger.error('Exception parsing and inserting file %s',
                             p, exc_info=True)
        logger.info('Successfully parsed and inserted %d of %d files for '
                    'account %d; %d files already in DB', success, len(files),
                    acct_id, already)

    def _do_one_file(self, acct_id, path):
        """
        Parse one OFX file and use OFXUpdater to upsert it into the DB.

        :param acct_id: Account ID number
        :type acct_id: int
        :param path: absolute path to OFX/QFX file
        :type path: str
        """
        logger.debug('Handle file %s for Account %d', path, acct_id)
        with open(path, 'rb') as fh:
            ofx_str = fh.read()
        ofx = OfxParser.parse(BytesIO(ofx_str))
        logger.debug('Parsed OFX')
        fname = os.path.basename(path)
        mtime = datetime.fromtimestamp(os.path.getmtime(path), tz=UTC)
        self._client.update_statement_ofx(
            acct_id, ofx, mtime=mtime, filename=fname
        )
        logger.debug('Done updating')


def parse_args():
    """
    Parse command-line arguments.
    """
    p = argparse.ArgumentParser(description='Backfill OFX from disk')
    p.add_argument('-v', '--verbose', dest='verbose', action='count', default=0,
                   help='verbose output. specify twice for debug-level output.')
    p.add_argument('-r', '--remote', dest='remote', action='store', type=str,
                   default=None,
                   help='biweeklybudget API URL to use instead of direct DB '
                        'access')
    p.add_argument('-s', '--save-path', dest='save_path', action='store',
                   type=str, default=None,
                   help='Statement save path; must be specified when running '
                        'in remote (-r) mode.')
    p.add_argument('--ca-bundle', dest='ca_bundle', action='store', type=str,
                   default=None,
                   help='Path to CA certificate bundle file or directory to '
                        'use for SSL verification')
    p.add_argument('--client-cert', dest='client_cert', action='store',
                   type=str, default=None,
                   help='path to client certificate to use for SSL client '
                        'cert auth')
    p.add_argument('--client-key', dest='client_key', action='store',
                   type=str, default=None,
                   help='path to unencrypted client key to use for SSL client '
                        'cert auth, if key is not contained in the cert file')
    args = p.parse_args()
    return args


def main():
    """
    Main entry point - instantiate and run :py:class:`~.OfxBackfiller`.
    """
    global logger
    format = "[%(asctime)s %(levelname)s] %(message)s"
    logging.basicConfig(level=logging.WARNING, format=format)
    logger = logging.getLogger()

    args = parse_args()

    # set logging level
    if args.verbose > 1:
        set_log_debug(logger)
    elif args.verbose == 1:
        set_log_info(logger)

    client = apiclient(
        api_url=args.remote, ca_bundle=args.ca_bundle,
        client_cert=args.client_cert, client_key=args.client_key
    )

    if args.remote is None:
        from biweeklybudget import settings
        save_path = settings.STATEMENTS_SAVE_PATH
    else:
        if args.save_path is None:
            logger.error('ERROR: -s|--save-path must be specified when running '
                         'in remote mode.')
            raise SystemExit(1)
        save_path = os.path.abspath(args.save_path)

    cls = OfxBackfiller(client, save_path)
    cls.run()


if __name__ == "__main__":
    main()



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
from datetime import datetime, timedelta, timezone
from typing import List
import pytest
import time
import re

from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.action_chains import ActionChains

from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models import (
    Account, Transaction, BudgetTransaction, PlaidAccount, PlaidItem,
    OFXTransaction, OFXStatement
)
from biweeklybudget.utils import dtnow
from plaid import Configuration, ApiClient, Environment
from plaid.api.plaid_api import PlaidApi
from plaid.models import SandboxItemResetLoginRequest

SANDBOX_USERNAME: str = 'user_good'
SANDBOX_PASSWORD: str = 'pass_good'

ONE_HOUR = timedelta(hours=1)


@pytest.mark.plaid
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
@pytest.mark.xfail(
    os.environ.get('CI') == 'true',
    reason="Plaid Link iframe flow incompatible with headless Chrome in CI. "
           "After clicking Continue on 'Your accounts' screen, Plaid's JavaScript "
           "fails to advance to the 'Save with Plaid' screen in headless mode, "
           "causing timeout waiting for 'Finish without saving' button. "
           "Flow works correctly in non-headless Chrome. "
           "See: https://github.com/jantman/biweeklybudget/pull/306#issuecomment-3757057657"
)
class TestLinkAndUpdateSimple(AcceptanceHelper):

    plaid_accts = {}
    plaid_acct_ids = []

    def test_00_clean_transactions_and_setup(self, testdb):
        for stmt in [
            'UPDATE accounts SET plaid_item_id=NULL, plaid_account_id=NULL;',
            'DELETE FROM plaid_accounts;',
            'DELETE FROM plaid_items;',
            'DELETE FROM txn_reconciles;',
            'DELETE FROM ofx_trans;',
            'DELETE FROM ofx_statements;',
            'DELETE FROM budget_transactions;',
            'DELETE FROM transactions;',
        ]:
            testdb.execute(stmt)
        testdb.flush()
        testdb.commit()
        self.plaid_accts = {}
        self.plaid_acct_ids = []

    def test_01_verify_db_state(self, testdb):
        for acct in testdb.query(Account).all():
            assert acct.plaid_account_id is None
            assert acct.plaid_item_id is None
        assert len(testdb.query(PlaidAccount).all()) == 0
        assert len(testdb.query(PlaidItem).all()) == 0
        assert len(testdb.query(Transaction).all()) == 0
        assert len(testdb.query(BudgetTransaction).all()) == 0
        assert len(testdb.query(OFXStatement).all()) == 0
        assert len(testdb.query(OFXTransaction).all()) == 0

    def test_02_plaid_items_table_is_empty(self, base_url, selenium):
        self.get(selenium, base_url + '/plaid-update')
        table = selenium.find_element(By.ID, 'table-items-plaid')
        texts = self.tbody2textlist(table)
        assert texts == []

    def test_03_do_simple_link(self, base_url, selenium):
        self.get(selenium, base_url + '/plaid-update')
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        selenium.find_element(By.ID, 'btn_link_plaid').click()
        self.wait_for_jquery_done(selenium)
        WebDriverWait(selenium, 30).until(
            EC.frame_to_be_available_and_switch_to_it(
                (By.ID, 'plaid-link-iframe-1')
            )
        )
        # inside iframe
        # click Continue as guest button
        WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable(
                (By.XPATH, '//button[normalize-space()="Continue as guest"]')
            )
        ).click()
        # search box - wait for it to be clickable
        sb = WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'search-input'))
        )
        # Use ActionChains to interact with the custom search input
        actions = ActionChains(selenium)
        actions.move_to_element(sb).click().send_keys('First Platypus').perform()
        # click on First Platypus Bank
        WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'aut-ins_109508'))
        ).click()
        # click on the non-OAuth institution
        WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'aut-ins_109508'))
        ).click()
        # enter username - use ActionChains for custom input fields
        user = WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'aut-input-0'))
        )
        actions = ActionChains(selenium)
        actions.move_to_element(user).click().send_keys(SANDBOX_USERNAME).perform()
        # enter password
        passwd = WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'aut-input-1'))
        )
        actions = ActionChains(selenium)
        actions.move_to_element(passwd).click().send_keys(SANDBOX_PASSWORD).perform()
        # click submit
        WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'aut-button'))
        ).click()
        # wait for accounts confirmation screen
        WebDriverWait(selenium, 10).until(
            EC.text_to_be_present_in_element(
                (By.CSS_SELECTOR, 'div.Content-module__content'),
                'Your accounts'
            )
        )
        # click continue
        WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'aut-button'))
        ).click()
        # Success message appears briefly, then Plaid may:
        # 1. Auto-advance to "Save with Plaid" screen (shows "Finish without saving" button), or
        # 2. Auto-close the iframe and complete successfully
        # Try to click "Finish without saving" button if it appears, otherwise continue
        try:
            WebDriverWait(selenium, 15).until(
                EC.element_to_be_clickable((By.ID, 'aut-secondary-button'))
            ).click()
            # Switch back to main content after closing iframe
            selenium.switch_to.default_content()
        except Exception:
            # Button didn't appear - Plaid may have auto-completed the flow
            # Try to switch back to main content (may already be there)
            try:
                selenium.switch_to.default_content()
            except Exception:
                pass
        time.sleep(2)  # Give the app time to process the Plaid callback
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        WebDriverWait(selenium, 10).until(
            EC.visibility_of_element_located((By.ID, 'table-items-plaid'))
        )
        WebDriverWait(selenium, 10).until(
            EC.presence_of_element_located((By.ID, 'table-items-plaid'))
        )
        WebDriverWait(selenium, 30).until(
            EC.text_to_be_present_in_element(
                (By.ID, 'table-items-plaid'),
                'First Platypus Bank'
            )
        )
        table = selenium.find_element(By.ID, 'table-items-plaid')
        texts = self.tbody2textlist(table)
        assert len(texts) == 1
        assert texts[0][1] == 'First Platypus Bank (ins_109508)'
        assert texts[0][2] == 'Plaid 401k (6666), ' \
                              'Plaid Business Credit Card (9999), ' \
                              'Plaid CD (2222), ' \
                              'Plaid Cash Management (9002), ' \
                              'Plaid Checking (0000), ' \
                              'Plaid Credit Card (3333), ' \
                              'Plaid HSA (9001), ' \
                              'Plaid IRA (5555), ' \
                              'Plaid Money Market (4444), ' \
                              'Plaid Mortgage (8888), ' \
                              'Plaid Saving (1111), ' \
                              'Plaid Student Loan (7777)'
        assert texts[0][3] in ['now', 'a second ago']

    def test_04_verify_db_updated(self, testdb):
        # check plaid items and accounts
        pitems: List[PlaidItem] = testdb.query(PlaidItem).all()
        assert len(pitems) == 1
        assert pitems[0].institution_name == 'First Platypus Bank'
        assert pitems[0].institution_id == 'ins_109508'
        assert dtnow() - pitems[0].last_updated < ONE_HOUR
        paccts: List[PlaidAccount] = testdb.query(PlaidAccount).all()
        assert len(paccts) == 12
        # find and check the checking and credit accounts
        checking = None
        credit = None
        for item in paccts:
            self.plaid_acct_ids.append(item.account_id)
            if item.name == 'Plaid Checking':
                checking = item
            elif item.name == 'Plaid Credit Card':
                credit = item
        assert checking is not None
        assert checking.mask == '0000'
        assert checking.account_type == 'depository'
        assert checking.account_subtype == 'checking'
        self.plaid_accts['checking'] = {
            'name': checking.name,
            'acct_id': checking.account_id,
            'item_id': checking.item_id
        }
        assert credit is not None
        assert credit.mask == '3333'
        assert credit.account_type == 'credit'
        assert credit.account_subtype == 'credit card'
        self.plaid_accts['credit'] = {
            'name': credit.name,
            'acct_id': credit.account_id,
            'item_id': credit.item_id
        }

    def test_05_associate_accounts_in_db(self, testdb):
        acct_chk: Account = testdb.query(Account).get(1)
        acct_chk.plaid_item_id = self.plaid_accts['checking']['item_id']
        acct_chk.plaid_account_id = self.plaid_accts['checking']['acct_id']
        testdb.add(acct_chk)
        acct_credit: Account = testdb.query(Account).get(3)
        acct_credit.plaid_item_id = self.plaid_accts['credit']['item_id']
        acct_credit.plaid_account_id = self.plaid_accts['credit']['acct_id']
        testdb.add(acct_credit)
        testdb.commit()

    def test_06_verify_no_transactions_in_db(self, testdb):
        assert len(testdb.query(Transaction).all()) == 0
        assert len(testdb.query(BudgetTransaction).all()) == 0
        assert len(testdb.query(OFXStatement).all()) == 0
        assert len(testdb.query(OFXTransaction).all()) == 0

    def test_07_update_transactions(self, base_url, selenium):
        self.get(selenium, base_url + '/plaid-update')
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        print('Waiting 30 seconds for item to be ready...')
        time.sleep(30)
        selenium.find_element(By.ID, 'btn_plaid_txns').click()
        self.wait_for_jquery_done(selenium)
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        WebDriverWait(selenium, 10).until(
            EC.visibility_of_element_located(
                (By.ID, 'table-accounts-plaid')
            )
        )
        div = selenium.find_element(By.CLASS_NAME, 'unreconciled-alert')
        assert re.match(r'\d{2} Unreconciled OFXTransactions\.', div.text)
        table = selenium.find_element(By.ID, 'table-accounts-plaid')
        texts = self.tbody2textlist(table)
        assert texts[0][0] == f"First Platypus Bank (" \
                              f"{self.plaid_accts['credit']['item_id']})"
        assert texts[0][1] == '0'
        assert 10 < int(texts[0][2]) < 20
        assert texts[0][3] == 'None'
        assert texts[0][4] == ''
        assert texts[1][0] == 'Total'

    def test_08_verify_new_transactions_in_db(self, testdb):
        assert len(testdb.query(Transaction).all()) == 0
        assert len(testdb.query(BudgetTransaction).all()) == 0
        assert len(testdb.query(OFXStatement).all()) == 2
        assert len(
            testdb.query(OFXStatement).filter(
                OFXStatement.account_id == 1
            ).all()
        ) == 1
        assert len(
            testdb.query(OFXStatement).filter(
                OFXStatement.account_id == 3
            ).all()
        ) == 1
        assert len(testdb.query(OFXTransaction).all()) > 10
        assert len(
            testdb.query(OFXTransaction).filter(
                OFXTransaction.account_id == 1
            ).all()
        ) > 5
        assert len(
            testdb.query(OFXTransaction).filter(
                OFXTransaction.account_id == 3
            ).all()
        ) > 5

    def test_09_update_transactions_again(self, base_url, selenium, testdb):
        orig_updated: datetime = testdb.query(PlaidItem).get(
            self.plaid_accts['checking']['item_id']
        ).last_updated
        assert len(testdb.query(OFXStatement).all()) == 2
        assert len(testdb.query(OFXTransaction).all()) > 10
        testdb.flush()
        testdb.commit()
        self.get(selenium, base_url + '/plaid-update')
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        selenium.find_element(By.ID, 'btn_plaid_txns').click()
        self.wait_for_jquery_done(selenium)
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        WebDriverWait(selenium, 10).until(
            EC.visibility_of_element_located(
                (By.ID, 'table-accounts-plaid')
            )
        )
        div = selenium.find_element(By.CLASS_NAME, 'unreconciled-alert')
        assert re.match(r'\d{2} Unreconciled OFXTransactions\.', div.text)
        table = selenium.find_element(By.ID, 'table-accounts-plaid')
        texts = self.tbody2textlist(table)
        assert texts[0][0] == f"First Platypus Bank (" \
                              f"{self.plaid_accts['credit']['item_id']})"
        assert 10 < int(texts[0][1]) < 20
        assert texts[0][2] == '0'
        assert texts[0][3] == 'None'
        assert texts[0][4] == ''
        assert texts[1][0] == 'Total'
        assert len(testdb.query(OFXStatement).all()) == 4
        assert len(testdb.query(OFXTransaction).all()) > 10
        new_updated: datetime = testdb.query(PlaidItem).get(
            self.plaid_accts['checking']['item_id']
        ).last_updated
        assert new_updated > orig_updated

    def test_10_change_item_institutions_in_db(self, testdb):
        pitems: List[PlaidItem] = testdb.query(PlaidItem).all()
        assert len(pitems) == 1
        pitems[0].institution_name = 'Wrong in DB'
        pitems[0].institution_id = 'wrongInDB'
        testdb.add(pitems[0])
        testdb.commit()

    def test_11_update_item_info(self, base_url, selenium):
        self.get(selenium, base_url + '/plaid-update')
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        button = selenium.find_element(By.ID, 'btn_import_plaid')
        button.click()
        # wait for page reload, indicated by staleness of element
        WebDriverWait(selenium, 20).until(
            EC.staleness_of(button)
        )

    def test_12_verify_item_info_updated(self, testdb):
        pitems: List[PlaidItem] = testdb.query(PlaidItem).all()
        assert len(pitems) == 1
        assert pitems[0].institution_name == 'First Platypus Bank'
        assert pitems[0].institution_id == 'ins_109508'

    def test_13_break_some_plaid_accounts(self, testdb):
        assert len(testdb.query(PlaidAccount).all()) == 12
        item_id = self.plaid_accts['checking']['item_id']
        chk_acct_id = self.plaid_accts['checking']['acct_id']
        credit_acct_id = self.plaid_accts['credit']['acct_id']
        acct_ids = [x for x in self.plaid_acct_ids]
        acct_ids.remove(chk_acct_id)
        acct_ids.remove(credit_acct_id)
        # remove/change two of the other accounts from the DB
        testdb.delete(testdb.query(PlaidAccount).get((item_id, acct_ids[0])))
        acct = testdb.query(PlaidAccount).get((item_id, acct_ids[-1]))
        acct.account_id = 'fooBarBaz'
        testdb.add(acct)
        testdb.flush()
        testdb.commit()
        all_accts: List[PlaidAccount] = testdb.query(PlaidAccount).all()
        assert len(all_accts) == 11
        assert 'fooBarBaz' in [x.account_id for x in all_accts]

    def test_14_refresh_plaid_accounts_for_item(self, base_url, selenium):
        self.get(selenium, base_url + '/plaid-update')
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        item_id = self.plaid_accts['checking']['item_id']
        button = selenium.find_element(By.CSS_SELECTOR,
                                       f'a[onclick="plaidRefresh(\'{item_id}\')"]'
                                       )
        button.click()
        # wait for page reload, indicated by staleness of element
        WebDriverWait(selenium, 20).until(
            EC.staleness_of(button)
        )

    def test_15_verify_accounts_updated_in_db(self, testdb):
        all_accts: List[PlaidAccount] = testdb.query(PlaidAccount).all()
        assert len(all_accts) == 12
        acct_ids = sorted([x.account_id for x in all_accts])
        assert acct_ids == sorted(self.plaid_acct_ids)

    def test_16_plaid_api_set_item_needs_login(self, testdb):
        pitem: PlaidItem = testdb.query(PlaidItem).all()[0]
        assert pitem.access_token is not None
        assert pitem.access_token != ''
        configuration = Configuration(
            host=getattr(Environment, os.environ['PLAID_ENV']),
            api_key={
                'clientId': os.environ['PLAID_CLIENT_ID'],
                'secret': os.environ['PLAID_SECRET'],
            }
        )
        api_client = ApiClient(configuration)
        client = PlaidApi(api_client)
        res = client.sandbox_item_reset_login(
            SandboxItemResetLoginRequest(access_token=pitem.access_token)
        )
        print(f'Reset login response: {res}')
        assert res.get('reset_login') is True

    def test_17_try_update_transactions_expect_error(self, base_url, selenium):
        self.get(selenium, base_url + '/plaid-update')
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        selenium.find_element(By.ID, 'btn_plaid_txns').click()
        self.wait_for_jquery_done(selenium)
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        WebDriverWait(selenium, 10).until(
            EC.visibility_of_element_located(
                (By.ID, 'table-accounts-plaid')
            )
        )
        table = selenium.find_element(By.ID, 'table-accounts-plaid')
        texts = self.tbody2textlist(table)
        assert texts[0][0] == f"First Platypus Bank (" \
                              f"{self.plaid_accts['credit']['item_id']})"
        assert texts[0][1] == '0'
        assert texts[0][2] == '0'
        assert 'the login details of this item have changed' in texts[0][3]
        assert texts[0][4] == ''
        assert texts[1][0] == 'Total'

    def test_18_update_item_relogin(self, base_url, selenium, testdb):
        item_id = self.plaid_accts['checking']['item_id']
        assert len(testdb.query(OFXStatement).all()) == 4
        num_txns = len(testdb.query(OFXTransaction).all())
        assert num_txns > 10
        testdb.flush()
        testdb.commit()
        self.get(selenium, base_url + '/plaid-update')
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        button = selenium.find_element(By.CSS_SELECTOR,
                                       f'a[onclick="plaidUpdate(\'{item_id}\')"]'
                                       )
        button.click()
        self.wait_for_jquery_done(selenium)
        WebDriverWait(selenium, 30).until(
            EC.frame_to_be_available_and_switch_to_it(
                (By.ID, 'plaid-link-iframe-1')
            )
        )
        # inside iframe
        # click Continue as guest button
        WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable(
                (By.XPATH, '//button[normalize-space()="Continue as guest"]')
            )
        ).click()
        # enter username - use ActionChains for custom input fields
        user = WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'aut-input-0'))
        )
        actions = ActionChains(selenium)
        actions.move_to_element(user).click().send_keys(SANDBOX_USERNAME).perform()
        # enter password
        passwd = WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'aut-input-1'))
        )
        actions = ActionChains(selenium)
        actions.move_to_element(passwd).click().send_keys(SANDBOX_PASSWORD).perform()
        # click submit
        WebDriverWait(selenium, 10).until(
            EC.element_to_be_clickable((By.ID, 'aut-button'))
        ).click()
        # For re-login flow, Plaid may show "Save" dialog or auto-close
        # Try to click "Finish without saving" button if it appears
        try:
            WebDriverWait(selenium, 5).until(
                EC.element_to_be_clickable((By.ID, 'aut-secondary-button'))
            ).click()
        except Exception:
            # Button didn't appear - Plaid may have auto-closed or flow changed
            # This is expected in some cases, continue with the test
            pass
        # Switch back to main content
        selenium.switch_to.default_content()
        time.sleep(2)  # Give the app time to process the Plaid callback
        # Handle success alert - use try/except in case it doesn't appear
        try:
            WebDriverWait(selenium, 10).until(EC.alert_is_present())
            selenium.switch_to.alert.accept()
        except Exception:
            # Alert didn't appear - this may happen if flow auto-completed
            pass
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        WebDriverWait(selenium, 10).until(
            EC.visibility_of_element_located((By.ID, 'table-items-plaid'))
        )
        WebDriverWait(selenium, 10).until(
            EC.presence_of_element_located((By.ID, 'table-items-plaid'))
        )
        WebDriverWait(selenium, 30).until(
            EC.text_to_be_present_in_element(
                (By.ID, 'table-items-plaid'),
                'First Platypus Bank'
            )
        )
        new_updated: datetime = testdb.query(PlaidItem).get(
            self.plaid_accts['checking']['item_id']
        ).last_updated

    def test_19_update_transactions(self, base_url, selenium, testdb):
        orig_updated: datetime = testdb.query(PlaidItem).get(
            self.plaid_accts['checking']['item_id']
        ).last_updated
        assert len(testdb.query(OFXStatement).all()) == 4
        num_txns = len(testdb.query(OFXTransaction).all())
        assert num_txns > 10
        testdb.flush()
        testdb.commit()
        self.get(selenium, base_url + '/plaid-update')
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        selenium.find_element(By.ID, 'btn_plaid_txns').click()
        self.wait_for_jquery_done(selenium)
        self.wait_for_load_complete(selenium)
        self.wait_for_jquery_done(selenium)
        WebDriverWait(selenium, 10).until(
            EC.visibility_of_element_located(
                (By.ID, 'table-accounts-plaid')
            )
        )
        div = selenium.find_element(By.CLASS_NAME, 'unreconciled-alert')
        assert re.match(r'\d{2} Unreconciled OFXTransactions\.', div.text)
        table = selenium.find_element(By.ID, 'table-accounts-plaid')
        texts = self.tbody2textlist(table)
        assert texts[0][0] == f"First Platypus Bank (" \
                              f"{self.plaid_accts['credit']['item_id']})"
        assert 10 < int(texts[0][1]) < 20
        assert texts[0][2] == '0'
        assert texts[0][3] == 'None'
        assert texts[0][4] == ''
        assert texts[1][0] == 'Total'
        assert len(testdb.query(OFXStatement).all()) == 6
        assert len(testdb.query(OFXTransaction).all()) == num_txns
        new_updated: datetime = testdb.query(PlaidItem).get(
            self.plaid_accts['checking']['item_id']
        ).last_updated
        assert new_updated > orig_updated



"""add BudgetTransaction model

Revision ID: 04e61490804b
Revises: 6d37400ea9cd
Create Date: 2018-01-12 18:43:14.983133

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from sqlalchemy import Column, Integer, ForeignKey, Numeric
import logging

logger = logging.getLogger(__name__)

Session = sessionmaker()

Base = declarative_base()

# revision identifiers, used by Alembic.
revision = '04e61490804b'
down_revision = '6d37400ea9cd'
branch_labels = None
depends_on = None


class Transaction(Base):

    __tablename__ = 'transactions'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: Actual amount of the transaction
    actual_amount = Column(Numeric(precision=10, scale=4), nullable=False)

    #: Budgeted amount of the transaction, if it was budgeted ahead of time
    #: via a :py:class:`~.ScheduledTransaction`.
    budgeted_amount = Column(Numeric(precision=10, scale=4))

    #: ID of the Budget this transaction is against
    budget_id = Column(Integer, ForeignKey('budgets.id'))

    #: ID of the Budget this transaction was planned to be funded by, if it
    #: was planned ahead via a :py:class:`~.ScheduledTransaction`
    planned_budget_id = Column(Integer, ForeignKey('budgets.id'))

    def __repr__(self):
        return "<Transaction(id=%s)>" % (
            self.id
        )


class BudgetTransaction(Base):
    """
    Represents the portion (amount) of a Transaction that is allocated
    against a specific budget. There will be one or more BudgetTransactions
    associated with each :py:class:`~.Transaction`.
    """

    __tablename__ = 'budget_transactions'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: Amount of the transaction against this budget
    amount = Column(Numeric(precision=10, scale=4), nullable=False)

    #: ID of the Transaction this is part of
    trans_id = Column(Integer, ForeignKey('transactions.id'))

    #: Relationship - the :py:class:`~.Transaction` this is part of
    transaction = relationship(
        "Transaction", backref="budget_transactions", uselist=False
    )

    #: ID of the Budget this transaction is against
    budget_id = Column(Integer, ForeignKey('budgets.id'))

    def __repr__(self):
        return "<BudgetTransaction(id=%s)>" % (
            self.id
        )


class Budget(Base):

    __tablename__ = 'budgets'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)


def upgrade():
    op.create_table(
        'budget_transactions',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('amount', sa.Numeric(precision=10, scale=4), nullable=False),
        sa.Column('trans_id', sa.Integer(), nullable=True),
        sa.Column('budget_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['budget_id'], ['budgets.id'],
            name=op.f('fk_budget_transactions_budget_id_budgets')
        ),
        sa.ForeignKeyConstraint(
            ['trans_id'], ['transactions.id'],
            name=op.f('fk_budget_transactions_trans_id_transactions')
        ),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_budget_transactions')),
        mysql_engine='InnoDB'
    )
    op.add_column(
        'transactions',
        sa.Column('planned_budget_id', sa.Integer(), nullable=True)
    )
    op.create_foreign_key(
        op.f('fk_transactions_planned_budget_id_budgets'),
        'transactions', 'budgets', ['planned_budget_id'], ['id']
    )
    # Copy amount and budget_id from each existing Transaction to a
    # corresponding BudgetTransaction
    bind = op.get_bind()
    session = Session(bind=bind)
    for txn in session.query(Transaction).all():
        b = BudgetTransaction(
            amount=txn.actual_amount,
            trans_id=txn.id,
            budget_id=txn.budget_id
        )
        session.add(b)
        if txn.budgeted_amount is not None:
            txn.planned_budget_id = txn.budget_id
            session.add(txn)
    session.commit()
    op.drop_constraint(
        'fk_transactions_budget_id_budgets', 'transactions', type_='foreignkey'
    )
    op.drop_column('transactions', 'budget_id')
    op.drop_column('transactions', 'actual_amount')


def downgrade():
    op.add_column(
        'transactions',
        sa.Column(
            'budget_id',
            sa.Integer(),
            autoincrement=False,
            nullable=True
        )
    )
    op.create_foreign_key(
        'fk_transactions_budget_id_budgets',
        'transactions', 'budgets', ['budget_id'], ['id']
    )
    op.add_column(
        'transactions',
        sa.Column(
            'actual_amount',
            Numeric(precision=10, scale=4),
            nullable=False
        )
    )
    # migrate budget_transactions back to Transaction.budget_id
    bind = op.get_bind()
    session = Session(bind=bind)
    for txn in session.query(Transaction).all():
        bt = txn.budget_transactions[0]
        txn.budget_id = bt.budget_id
        txn.actual_amount = bt.amount
        session.add(txn)
    session.commit()
    op.drop_table('budget_transactions')
    op.drop_constraint(
        op.f('fk_transactions_planned_budget_id_budgets'),
        'transactions',
        type_='foreignkey'
    )
    op.drop_column('transactions', 'planned_budget_id')
